from copy import deepcopy
from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
from collections import defaultdict
import itertools
import logging
import time
from typing import Any, Callable

from accelforge._accelerated_imports import pd, np
from accelforge.frontend.spec import Spec
from accelforge.frontend.mapping import Mapping
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.frontend.workload import EinsumName
from accelforge.mapper.FFM.mappings import Mappings
from accelforge.mapper.FFM.pmappings import MultiEinsumPmappings
from accelforge.mapper.FFM._join_pmappings.compress_pmappings import (
    compress_einsum2pmappings,
    decompress_pmappings,
)
from accelforge.mapper.FFM._make_pmappings.make_pmappings import (
    get_rank_variable_bounds_for_all_einsums,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import (
    row2pmappings,
)
from accelforge.mapper.FFM._pareto_df.df_convention import (
    MAPPING_COLUMN,
    col_used_in_pareto,
    is_objective_col,
    is_reservation_col,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_group import (
    PmappingGroup,
    Compatibility,
)
from accelforge.mapper.FFM._pareto_df.df_convention import col2reservation
from accelforge.util import _fillna_and__numeric_cast, parallel, delayed, oset

logger = logging.getLogger(__name__)


class JoiningTimer:
    def __init__(self):
        self.prev_time = time.time()
        self.total_time = defaultdict(int)

    def print_time(self, what: str):
        t = time.time() - self.prev_time
        logger.info(f"{what}: {t:.2f} seconds")
        self.total_time[what] += t
        self.prev_time = time.time()

    def log_total_time(self):
        logger.info(f"\n======== Total time ========")
        for k, v in self.total_time.items():
            logger.info(f"{k}: {v:.2f} seconds")
        total = sum(self.total_time.values())
        if total > 60:
            logger.info(f"\nTotal: {total:.2f} seconds ({total/60:.2f} minutes)")
        else:
            logger.info(f"\nTotal: {total:.2f} seconds")
        logger.info(f"============================\n")


class OptimalityThresholder:
    def __init__(
        self,
        prev_solutions: Mappings,
        _pmapping_row_filter_function: Callable[[pd.DataFrame], np.ndarray],
        aggregator: str,
        print_progress: bool,
    ):
        compare_to = prev_solutions.data
        compare_cols = [c for c in compare_to.columns if col_used_in_pareto(c)]
        self._pmapping_row_filter_function = _pmapping_row_filter_function
        self.aggregator = aggregator

        if self.aggregator in ("prod", "sum"):
            objective_cols = [c for c in compare_cols if is_objective_col(c)]
            self._agg_cols = objective_cols
            if objective_cols:
                values = np.column_stack(
                    [compare_to[c].values for c in objective_cols]
                )
                if self.aggregator == "prod":
                    agg = np.prod(values, axis=1)
                else:
                    agg = np.sum(values, axis=1)
                self._agg_threshold = agg.min()
            else:
                self._agg_threshold = float("inf")
            if print_progress:
                label = "product" if self.aggregator == "prod" else "sum"
                print(
                    f"Filtering out pmappings with {label} > "
                    f"{self._agg_threshold:.2e}"
                )
        else:  # "any"
            compare_to = compare_to.sort_values(by=compare_cols, ascending=False)

            if len(compare_to) > 10:
                chosen_indices = np.round(
                    np.linspace(0, len(compare_to) - 1, 10)
                )
            else:
                chosen_indices = np.round(np.arange(len(compare_to)))

            self.compare_to: list[dict[str, float]] = []
            if print_progress:
                print(f"Filtering out pmappings worse than the following:")

            for i in chosen_indices.astype(int):
                self.compare_to.append(
                    {c: compare_to[c].iloc[i] for c in compare_cols}
                )
                if print_progress:
                    print(
                        "\t"
                        + "    ".join(
                            f"{k}={float(v):.2e}"
                            for k, v in self.compare_to[-1].items()
                        )
                    )

    def __call__(self, mapping: pd.DataFrame) -> bool:
        nondominated_by_all = np.ones(len(mapping), dtype=bool)

        if self.aggregator in ("prod", "sum"):
            cols_present = [c for c in self._agg_cols if c in mapping.columns]
            if cols_present:
                values = np.column_stack(
                    [mapping[c].values for c in cols_present]
                )
                if self.aggregator == "prod":
                    agg = np.prod(values, axis=1)
                else:
                    agg = np.sum(values, axis=1)
                nondominated_by_all = agg <= self._agg_threshold
        else:  # "any"
            for c in self.compare_to:
                nondominated = np.zeros(len(mapping), dtype=bool)
                for k, v in c.items():
                    if k not in mapping.columns:
                        nondominated |= True
                    else:
                        nondominated |= mapping[k] <= v
                nondominated_by_all &= nondominated

        if self._pmapping_row_filter_function is not None:
            nondominated_by_all &= self._pmapping_row_filter_function(mapping)

        return nondominated_by_all


def prune_with_tolerance(
    pmappings: dict[EinsumName, list[PmappingGroup]],
    objective_tolerance: float,
    resource_usage_tolerance: float,
    print_progress: bool = True,
):
    if objective_tolerance == 0 and resource_usage_tolerance == 0:
        return pmappings

    prev_n = sum(len(pg.mappings) for p in pmappings.values() for pg in p)

    def prune(einsum_name: EinsumName, pg: PmappingGroup):
        pg = PmappingGroup(pg.compatibility, pg.mappings.make_pareto(
            objective_tolerance=objective_tolerance,
            resource_usage_tolerance=resource_usage_tolerance,
            inplace=False,
        ))
        return einsum_name, pg

    jobs = [delayed(prune)(e, pg) for e, p in pmappings.items() for pg in p]

    result = {einsum_name: [] for einsum_name in pmappings.keys()}
    for einsum_name, pg in parallel(
        jobs, pbar="Dirty pruning pmappings" if print_progress else None
    ):
        result[einsum_name].append(pg)

    new_n = sum(len(pg.mappings) for p in result.values() for pg in p)
    if print_progress:
        print(f"Dirty joining uses {new_n / prev_n * 100:.2f}% of the pmappings")

    return result


def join_strategy_2(
    spec: Spec,
    compressed: dict[EinsumName, list[PmappingGroup]],
    print_progress: bool,
    metrics: Metrics,
    for_model: bool,
    _pmapping_row_filter_function: Callable[[pd.DataFrame], np.ndarray] | None = None,
    resource_usage_tolerance: float = 0,
):
    thresholds = [1, 0]
    thresholds = [t for t in thresholds if t > spec.mapper.objective_tolerance]
    thresholds.append(spec.mapper.objective_tolerance)

    filter_func = _pmapping_row_filter_function
    _runtime_log_file = spec.mapper._runtime_log_file
    for i, threshold in enumerate(thresholds):
        is_dirty = i < len(thresholds) - 1
        if not for_model and print_progress:
            if is_dirty:
                print(f"Dirty joining with objectives <= {1 + threshold}× optimal")
            else:
                print("Final clean join.")
        # Write round marker so the notebook can distinguish dirty vs clean
        if _runtime_log_file and is_dirty:
            import json
            with open(_runtime_log_file, "a") as f:
                f.write(json.dumps({"round": i, "threshold": threshold}) + "\n")
        try:
            cur_compressed = prune_with_tolerance(
                compressed,
                objective_tolerance=threshold,
                resource_usage_tolerance=resource_usage_tolerance,
                print_progress=print_progress,
            )
            joined = join_pmappings(
                cur_compressed,
                spec,
                _pmapping_row_filter_function=filter_func,
                print_progress=print_progress,
                metrics=metrics,
            )
            if i < len(thresholds) - 1:
                filter_func = OptimalityThresholder(
                    joined,
                    _pmapping_row_filter_function,
                    spec.mapper._metric_aggregator,
                    print_progress
                )
        except Exception as e:
            if i == len(thresholds) - 1:
                raise
            if print_progress:
                print(f"Error with optimality threshold {threshold}: {e}")

    return joined


def multi_strategy_join(
    spec: Spec,
    compressed: dict[EinsumName, list[PmappingGroup]],
    print_progress: bool,
    metrics: Metrics,
    for_model: bool,
    _pmapping_row_filter_function: Callable[[pd.DataFrame], np.ndarray] | None = None,
):
    for _, p in compressed.items():
        for pg in p:
            pg.mappings.drop_valid_reservations = not (Metrics.RESOURCE_USAGE & metrics)

    # If it's for the model, just join things directly
    if for_model:
        return join_pmappings(
            deepcopy(compressed),
            spec,
            print_progress=print_progress,
            metrics=metrics,
            _pmapping_row_filter_function=_pmapping_row_filter_function,
        )

    if metrics & Metrics.RESOURCE_USAGE:
        return join_strategy_2(
            spec,
            compressed,
            print_progress,
            metrics,
            for_model,
            _pmapping_row_filter_function,
        )

    resource_usage_thresholds = [0.02, 0.01, 0.001, 0.0001, 0]
    for i, threshold in enumerate(resource_usage_thresholds):
        for p in compressed.values():
            for pg in p:
                pg.mappings.excess_resource_tolerance = threshold
        if i < len(resource_usage_thresholds) - 1 and print_progress:
            print(f"Dirty joining with resource usage <= {1 + threshold}× optimal")
        joined = join_strategy_2(
            spec,
            compressed,
            print_progress,
            metrics,
            for_model,
            _pmapping_row_filter_function,
            resource_usage_tolerance=threshold,
        )
        for c in joined.data.columns:
            if is_reservation_col(c):
                maxvalue = joined.data[c].max()
                if maxvalue > 1:
                    if print_progress:
                        oversubscribed = (
                            f"{col2reservation(c).name} ({maxvalue * 100:.2f}%)"
                        )
                        print(f"Oversubscribed {oversubscribed}. Reducing threshold...")
                    break
        else:
            if print_progress:
                print("Dirty joining mapping(s) valid & optimal! Returning...")
            return joined
    return joined


def clean_compress_and_join_pmappings(
    pmappings: MultiEinsumPmappings,
    metrics: Metrics,
    for_model: bool,
    require_all_einsums: bool = True,
    _pmapping_row_filter_function: Callable[[pd.Series], bool] | None = None,
    print_progress: bool = True,
) -> Mappings:
    einsum2pmappings = pmappings.einsum2pmappings
    if not require_all_einsums:
        einsum2pmappings = {
            k: v
            for k, v in pmappings.einsum2pmappings.items()
            if k in pmappings.einsums_with_pmappings_generated
        }
    _check_einsum2pmappings_not_empty(einsum2pmappings, pmappings)

    compressed, decompress_data = compress_einsum2pmappings(
        einsum2pmappings, print_progress
    )

    joined = multi_strategy_join(
        pmappings.spec,
        compressed,
        print_progress,
        metrics,
        for_model,
        _pmapping_row_filter_function,
    )

    joined = decompress_pmappings(joined, decompress_data)

    for einsum_name in einsum2pmappings:
        col = f"{einsum_name}<SEP>{MAPPING_COLUMN}"
        joined.data[col] = joined.data[col].apply(
            lambda x: pmappings.pmapping_objects[einsum_name][x]
        )
    joined._data = _fillna_and__numeric_cast(joined.data, 0).reset_index(drop=True)

    rank_variable_bounds = get_rank_variable_bounds_for_all_einsums(pmappings.spec)
    einsum_names = list(einsum2pmappings.keys())
    joined.data[f"Total<SEP>{MAPPING_COLUMN}"] = [
        MappingFromRow(r, rank_variable_bounds, einsum_names)
        for _, r in joined.data.iterrows()
    ]
    # Fill nans with 0. We might get missing columns for some mapping entries if there
    # are energy entries for some pmappings but not others (e.g., one pmapping accesses
    # DRAM while another doesn't.)
    return Mappings(
        pmappings.spec,
        list(
            x
            for x in list(einsum2pmappings.keys())
            if x in pmappings.einsums_with_pmappings_generated
        ),
        joined.data,
        total_mappings=joined.n_total_pmappings,
        valid_mappings=joined.n_valid_pmappings,
        flattened_arches=pmappings.flattened_arches,
        evaluated_specs=pmappings.evaluated_specs,
    )


class PmappingsOneEinsum:
    def __init__(self, einsum_name: str, pm_group_list: list[PmappingGroup]):
        self.einsum_name: str = einsum_name
        self.pmapping_groups: list[PmappingGroup] = pm_group_list
        self.tensor_names: set[str] = oset(pm_group_list[0].tensor_names)

    def __getitem__(self, i):
        return self.pmapping_groups[i]


def make_full_equivalent_rank_variables(pairwise_equivalent_rank_variables):
    full_equivalent_rank_variables = {
        k: oset(v) for k, v in pairwise_equivalent_rank_variables.items()
    }
    changed = True
    while changed:
        changed = False
        for r in full_equivalent_rank_variables:
            for r2 in list(full_equivalent_rank_variables[r]):
                for r3 in list(full_equivalent_rank_variables[r2]):
                    if r3 in full_equivalent_rank_variables[r]:
                        continue
                    changed = True
                    full_equivalent_rank_variables[r].add(r3)
    return full_equivalent_rank_variables


def get_memories_to_track(
    pmapping_groups: dict[str, list[PmappingGroup]],
    print_progress: bool = True,
) -> tuple[dict[str, list[PmappingGroup]], set[str], set[str]]:

    always_below = oset()
    for _, einsum_pmapping_groups in pmapping_groups.items():
        for s in einsum_pmapping_groups:
            for col in s.mappings.data.columns:
                reservation_key = col2reservation(col)
                if reservation_key is not None:
                    always_below.add(reservation_key.name)

    total_sizes = {}
    ignored_resources = oset()

    for _, einsum_pmapping_groups in pmapping_groups.items():
        max_sizes = {}
        for s in einsum_pmapping_groups:
            n_fused_loops = s.compatibility.n_loops
            for col in s.mappings.data.columns:
                reservation_key = col2reservation(col)
                if reservation_key is None:
                    continue

                name = reservation_key.name
                nloops = reservation_key.nloops
                if name in always_below and nloops < n_fused_loops:
                    always_below.remove(name)
                # Check each of the compatibility's tensors
                for tensor in s.compatibility.tensors:
                    if tensor.resource_name in always_below:
                        always_below.remove(tensor.resource_name)
                size = s.mappings.data[col].max()
                max_sizes[name] = max(max_sizes.get(name, 0), size)

                # nloops < 0 means that the reservation will live through all Einsums
                if nloops < 0:
                    ignored_resources.add(name)

        for name, size in max_sizes.items():
            total_sizes[name] = total_sizes.get(name, 0) + size

    ignore = oset(t for t, s in total_sizes.items() if s <= 1) | always_below

    if not ignore:
        return pmapping_groups, ignore

    def remove_unneeded_columns(s: PmappingGroup):
        data = s.mappings.data
        keep_cols = []
        for col in data.columns:
            name_nloops = col2reservation(col)
            if name_nloops is None or name_nloops[0] not in ignore:
                keep_cols.append(col)
        run_pareto = len(keep_cols) < len(data.columns)
        data = data[keep_cols].copy() if len(keep_cols) < len(data.columns) else data
        return PmappingGroup(
            s.compatibility,
            s.mappings.update(data=data, skip_pareto=not run_pareto),
        )

    for a in sorted(always_below):
        if print_progress:
            print(
                f"Not tracking {a} because it is never reserved for multiple pmappings."
            )
    for t, s in sorted(total_sizes.items(), key=lambda x: x[1], reverse=True):
        if s <= 1:
            if print_progress:
                print(
                    f"Not tracking {t} because its size is enough for the sum of all "
                    f"reservations ({s * 100:.2f}% of the total)"
                )
            break

    new_pmapping_groups = {}
    for einsum_name, einsum_pmapping_groups in pmapping_groups.items():
        new_pmapping_groups[einsum_name] = parallel(
            [delayed(remove_unneeded_columns)(s) for s in einsum_pmapping_groups],
            pbar=(
                f"Removing unneeded reservations for {einsum_name}"
                if print_progress
                else None
            ),
        )
    return new_pmapping_groups, ignore


def join_pmappings(
    pmapping_groups: dict[str, list[PmappingGroup]],
    spec: Spec,
    lookahead_filter: bool = True,
    metrics: Metrics = None,
    _pmapping_row_filter_function: Callable[[pd.Series], bool] | None = None,
    print_progress: bool = True,
):
    """
    CONTRACT FOR MAPPINGS GETTING TO THIS POINT:

    - Reservations at a level include reservations at all levels above it.
    - If one Einsum uses an aliased tensor more than once, then only one
      reservation is made for it. If overlapping lifetimes cause the aliases to
      be alive at the same time, then it is handled here.
    - Memory names should be sorted with higher memory names representing
      memories lower in the hierarchy. e.g., memory 0 is the largest,
      memory 1 the next largest, and memory N is the smallest.
    """
    skip_invalid = spec.mapper._skip_invalid
    combine_reservations = spec.mapper._combine_reservations
    _runtime_log_file = spec.mapper._runtime_log_file

    drop_valid_reservations = not (Metrics.RESOURCE_USAGE & metrics)
    ignored_resources = oset()

    if _pmapping_row_filter_function is not None:
        n = sum(len(s.mappings.data) for sg in pmapping_groups.values() for s in sg)
        pmapping_groups = {
            e: [
                PmappingGroup(
                    s.compatibility,
                    s.mappings.filter_rows(_pmapping_row_filter_function),
                )
                for s in pmapping_groups[e]
            ]
            for e in pmapping_groups
        }
        new_n = sum(len(s.mappings.data) for sg in pmapping_groups.values() for s in sg)
        if print_progress:
            print(f"Filtered {n} -> {new_n} ({new_n / n:.2%} kept) pmappings")

    if drop_valid_reservations:
        pmapping_groups, ignored_resources = get_memories_to_track(
            pmapping_groups, print_progress
        )

    for einsum_name, einsum_pmapping_groups in pmapping_groups.items():
        for s in einsum_pmapping_groups:
            s.mappings.drop_valid_reservations = drop_valid_reservations

    aliased_tensors = spec.workload.get_tensor_copies()

    n_mappings = {}
    runtime = {}
    nbuckets = []

    n_evaluations = 0

    pmapping_groups = list(pmapping_groups.items())

    if not skip_invalid:
        lookahead_filter = False

    for einsum_name, s in pmapping_groups:
        if not s:
            raise ValueError(f"No pmappings for {einsum_name}")

    timer = JoiningTimer()

    pmgroups = [PmappingsOneEinsum(*s) for s in pmapping_groups]

    if not pmgroups:
        raise ValueError("No pmappings to join")

    # ======================================================================
    # Initial consolidate and group all PmappingGroups
    # ======================================================================
    n_mappings["Post Intra-Layer"] = 0
    for i, einsum_pmappings in enumerate(pmgroups):
        cur_tensors = einsum_pmappings.tensor_names
        right_tensors = oset.union(oset(), *[s.tensor_names for s in pmgroups[i + 1 :]])
        # First Einsum: Remove dead tensors and left consolidate. This is because the
        # first Einsum will have the first pmappigns that are joined from the left
        if i == 0:
            if cur_tensors - right_tensors:
                PmappingGroup.remove_dead_tensors(
                    einsum_pmappings.pmapping_groups, right_tensors
                )
                for s in einsum_pmappings.pmapping_groups:
                    s.compatibility = s.compatibility.clear_dead_tensors(right_tensors)
            einsum_pmappings.pmapping_groups = PmappingGroup.left_consolidate(
                einsum_pmappings.pmapping_groups,
                right_tensors,
                parallelize=False,  # We're not pareto pruning, so parallelization doesn't help.
                pbar=(
                    f"Inital consolidate {einsum_pmappings.einsum_name} ({i+1}/{len(pmgroups)})"
                    if print_progress
                    else None
                ),
            )
            continue

        # All other Einsums: Will be joined from the right. Remove dead tensors, right
        # consolidate, combine, group.
        t0 = time.time()
        left_tensors = oset.union(oset(), *[s.tensor_names for s in pmgroups[:i]])
        live_tensors = right_tensors
        shared_tensors = left_tensors & einsum_pmappings.tensor_names

        if cur_tensors - (right_tensors | left_tensors):
            PmappingGroup.remove_dead_tensors(
                einsum_pmappings.pmapping_groups, right_tensors | left_tensors
            )
            for s in einsum_pmappings.pmapping_groups:
                s.compatibility = s.compatibility.clear_dead_tensors(
                    right_tensors | left_tensors
                )

        einsum_pmappings.pmapping_groups = sorted(
            einsum_pmappings.pmapping_groups,
            key=lambda x: len(x.mappings.data),
            reverse=True,
        )
        einsum_pmappings.pmapping_groups = PmappingGroup.right_consolidate(
            einsum_pmappings.pmapping_groups,
            live_tensors,
            shared_tensors,
            parallelize=False,  # We're not pareto pruning, so parallelization doesn't help.
            pbar=(
                f"Inital consolidate {einsum_pmappings.einsum_name} ({i+1}/{len(pmgroups)})"
                if print_progress
                else None
            ),
        )
        einsum_pmappings.pmapping_groups = PmappingGroup.combine_combineable(
            einsum_pmappings.pmapping_groups,
            left_tensors | right_tensors,
            _combine_reservations=combine_reservations,
            pbar_postfix=f" for {einsum_pmappings.einsum_name} ({i+1}/{len(pmgroups)})",
            print_progress=print_progress,
        )
        n_mappings["Post Intra-Layer"] += sum(
            len(s.mappings.data) for s in einsum_pmappings.pmapping_groups
        )
        einsum_pmappings.pmapping_groups = PmappingGroup.group(
            einsum_pmappings.pmapping_groups, left_tensors
        )
        einsum, prev_einsum = einsum_pmappings.einsum_name, pmgroups[i - 1].einsum_name
        step_time = time.time() - t0
        runtime[f"{prev_einsum} → {einsum}"] = step_time
        if _runtime_log_file:
            import json as _json
            with open(_runtime_log_file, "a") as _f:
                _f.write(_json.dumps({"step": f"{prev_einsum} → {einsum}", "phase": "consolidate", "time": step_time}) + "\n")
        t0 = time.time()
    timer.print_time(f"Initial consolidate and group")

    n_iterations = 0
    total_iterations = len(pmgroups)

    def grab_einsum_pmappings() -> (
        tuple[dict[Compatibility, list[PmappingGroup]], str, set[str]]
    ):
        nonlocal n_iterations
        n_iterations += 1
        holder = pmgroups.pop(0)
        return holder.pmapping_groups, holder.einsum_name, holder.tensor_names

    if pmgroups:
        left, left_einsum, left_tensors = grab_einsum_pmappings()

    partial_mapping_size = 1
    while pmgroups:
        t0 = time.time()
        # ======================================================================
        # Check that data dependencies are satisfied.
        # ======================================================================
        for s in pmgroups:
            output_tensors = spec.workload.einsums[s.einsum_name].output_tensor_names
            shared_fail = left_tensors & output_tensors
            if shared_fail:
                raise ValueError(
                    f"Einsum {left_einsum} uses tensors {sorted(shared_fail)} that "
                    f"are outputs of Einsum {s.einsum_name}, which is later in the "
                    f"joining order."
                )

        # ======================================================================
        # Grab new Einsum from the right. Record logging data and find still
        # tensors that will be live after this Einsum.
        # ======================================================================
        nbuckets.append(len(left))
        # nmappings.append(sum(len(s.mappings.data) for s in left))
        right, right_einsum, right_tensors = grab_einsum_pmappings()
        logger.info(f"Einsum {right_einsum} ({n_iterations}/{total_iterations})")

        partial_mapping_size += 1

        live_tensors = oset.union(oset(), *[s.tensor_names for s in pmgroups])
        shared_tensors = oset(left_tensors) & oset(right_tensors)
        live_tensors_with_right = live_tensors | right_tensors

        # ======================================================================
        # Clean up the previously-combined PmappingGroups. Consolidate, combine, group
        # them into buckets.
        # ======================================================================
        # print_time(f"Consolidating")

        left = PmappingGroup.combine_combineable(
            left,
            live_tensors | right_tensors,
            _combine_reservations=combine_reservations,
            print_progress=print_progress,
        )

        # print_time(f"Combining")
        # Group left and right into buckets
        left = PmappingGroup.group(left, right_tensors)
        # print_time("Grouping")

        # ======================================================================
        # Remove dead tensors from left and right. This happens after grouping because
        # we only reserve space for shared tensors after they're dead (alive is handled
        # by the normal reservation system). This is in case the tensor lifetime extends
        # beyond the Einsums for which it is used.
        # ======================================================================
        PmappingGroup.remove_dead_tensors(
            [s for lr in [left, right] for v in lr.values() for s, _ in v], live_tensors
        )

        DO_PRINT = False
        DELAY = True
        # ======================================================================
        # Merge the left and right buckets.
        # ======================================================================
        combined: list[PmappingGroup] = []
        cur_nmappings = 0
        combined_ids: set[tuple[int, int, tuple[tuple[int, int], ...]]] = oset()

        for k in left:
            found = False
            if DO_PRINT:
                print(f"Left key {k}")
            for (a, perm_a), (b, perm_b) in itertools.product(
                left[k], right.get(k, [])
            ):
                a: PmappingGroup
                b: PmappingGroup
                perm_a: list[int]
                perm_b: list[int]
                key_check = (id(a), id(b))
                if key_check in combined_ids:
                    continue
                combined_ids.add(key_check)
                found = True

                compatibility_a = a.compatibility.permute(perm_a)
                compatibility_b = b.compatibility.permute(perm_b)
                try:
                    compatibility_joined = compatibility_a.merge_next(
                        compatibility_b,
                        live_tensors,
                    )
                    if DO_PRINT:
                        print(
                            f"\t{a.compatibility}        <-->        {b.compatibility}"
                        )
                except ValueError as e:  # Incompatible!
                    # if DO_PRINT:
                    #     print(f"\tIncompatible: {e}")
                    continue

                t0 = time.time()

                combined.append(
                    a.merge_next(
                        b,
                        live_tensors,
                        live_tensors_with_right,
                        aliased_tensors,
                        compatibility_joined=compatibility_joined,
                        permuted_compatibility_left=compatibility_a,
                        permuted_compatibility_right=compatibility_b,
                        delay=DELAY,
                        _pmapping_row_filter_function=_pmapping_row_filter_function,
                        ignored_resources=ignored_resources,
                    )
                )

                if not DELAY:
                    cur_nmappings += len(a.mappings.data) * len(b.mappings.data)
                if DO_PRINT:
                    # s = f"\t-->\n\t{combined[-1].compatibility}"
                    # s += f"({len(a.mappings.data)})x({len(b.mappings.data)})"
                    # print(s)
                    pass
            if DO_PRINT and not found:
                for a, _ in left[k]:
                    print(f"\tNo match for {a.compatibility}")

        if DO_PRINT:
            for k in right:
                if k not in left:
                    for b, _ in right[k]:
                        print(f"\tREVERSE: No match for {b.compatibility} using {k}")

        # Count before clearing (needed for skip_invalid=False)
        if not skip_invalid:
            _left_n = sum(len(s.mappings.data) for k in left.values() for s, _ in k)
            _right_n = sum(len(s.mappings.data) for k in right.values() for s, _ in k)

        for l in left.values():
            for s, _ in l:
                s.mappings = None
        for r in right.values():
            for s, _ in r:
                s.mappings = None

        # print_time("Bucket merging")
        def raise_no_match_error():
            estr = f"No match found for any group.\n"
            estr += f"Left compatibility:\n\t" + "\n\t".join(
                str(c) for c in left.keys()
            )
            estr += f"\nRight compatibility:\n\t" + "\n\t".join(
                str(c) for c in right.keys()
            )
            raise ValueError(estr)

        def no_match_lookahead_error(
            combined: list[PmappingGroup],
            next_keys: set[tuple[int, int, tuple[tuple[int, int], ...]]],
        ):
            estr = f"No match found for any group. Left and right joined successfully, "
            estr += f"but will not be compatible with following Einsums.\n"
            estr += f"Left compatibility:\n\t" + "\n\t".join(
                str(s.compatibility) for g in left.values() for s, _ in g
            )
            estr += f"\nRight compatibility:\n\t" + "\n\t".join(
                str(s.compatibility) for g in right.values() for s, _ in g
            )
            estr += f"\nCombined compatibility:\n\t" + "\n\t".join(
                str(s.compatibility) for s in combined
            )
            estr += f"\nFollowing Einsum compatibility:\n\t" + "\n\t".join(
                str(c) for c in next_keys
            )
            raise ValueError(estr)

        # ======================================================================
        # Look ahead to the next Einsum and see if any of our groups will not
        # be able to merge with it. If so, we can drop them immediately.
        # ======================================================================
        lookahead_filter = True
        if lookahead_filter:
            cur_tensors = left_tensors | right_tensors
            for next_pmapping_groups in pmgroups:
                next_right_tensors = next_pmapping_groups.tensor_names
                if not next_right_tensors & cur_tensors:
                    continue
                prev_combined = combined
                combined = PmappingGroup.group(combined, next_right_tensors)
                next_keys = oset(
                    c.clear_dead_tensors(
                        cur_tensors
                    ).clear_tile_patterns_and_reservation_indices()
                    for c in next_pmapping_groups.pmapping_groups
                )
                for k in list[PmappingGroup](combined):
                    perms = k.make_equivalent_permutations()
                    perms = [
                        p[0]
                        .clear_dead_tensors(next_right_tensors)
                        .clear_tile_patterns_and_reservation_indices()
                        for p in perms
                    ]
                    if not any(p in next_keys for p in perms):
                        if DO_PRINT:
                            for b, _ in combined[k]:
                                print(
                                    f"\tLOOKAHEAD to {next_pmapping_groups.einsum_name}: No match for {b.compatibility}"
                                )
                        del combined[k]
                if not combined:
                    PmappingGroup.group(prev_combined, next_right_tensors)
                    no_match_lookahead_error(prev_combined, next_keys)

                combined = list(itertools.chain.from_iterable(combined.values()))
                combined = [c[0] for c in combined]
                # Remove duplicates
                id2combined = {id(c): c for c in combined}
                combined = list(id2combined.values())
                # print(
                #     f"Removed {prev_len - len(combined)}/{prev_len} ({len(combined)/prev_len*100:.2f}% remaining)"
                # )
                # print_time("Removing mappings that can't be combined later")

        if not combined:
            raise_no_match_error()

        # ======================================================================
        # If we delayed the mapping merging, do it now.
        # ======================================================================
        import copy

        if DELAY:
            mappings = parallel(
                [c.mappings for c in combined],
                pbar=(
                    f"Joining pmappings for {left_einsum} <--> {right_einsum} ({n_iterations}/{total_iterations})"
                    if print_progress
                    else None
                ),
            )
            for c, mapping in zip(combined, mappings):
                c.mappings = mapping
                cur_nmappings += c.n_pre_prune_mappings
        timer.print_time("Pmapping merging")

        if not any(len(s.mappings.data) for s in combined):
            # for c in prev_combined:  # For debugging the joining
            #     x = c.mappings
            #     x[0](*x[1], **x[2])
            raise ValueError(f"No mappings found for {left_einsum} <--> {right_einsum}")

        prev_nmappings = cur_nmappings
        if not skip_invalid:
            cur_nmappings = _left_n * _right_n
        n_mappings[f"{left_einsum} → {right_einsum}"] = cur_nmappings
        n_evaluations += cur_nmappings
        step_time = (time.time() - t0) * (cur_nmappings / prev_nmappings)
        runtime[f"{left_einsum} → {right_einsum}"] += step_time
        if _runtime_log_file:
            import json as _json
            with open(_runtime_log_file, "a") as _f:
                _f.write(_json.dumps({"step": f"{left_einsum} → {right_einsum}", "phase": "join", "time": step_time}) + "\n")
        # print(
        #     f'Scaled runtime by {cur_nmappings / prev_nmappings}. Runtime: {runtime[f"{prev_einsum} → {einsum}"]:.2f}'
        # )

        # # ======================================================================
        # # Print statements
        # # ======================================================================
        # logger.info(
        #     f"\tCombining {sum(len(s) for s in left.values())}({len(left)}) x {sum(len(s) for s in right.values())}({len(right)}) -> {len(combined)}"
        # )

        # nmappings = sum(len(s.mappings.data) for s in combined)
        # for_einsum_text = f"for Einsum {right_einsum}"
        # logger.info(f"\tNumber of groups {for_einsum_text}: {len(combined)}")
        # # for c in combined:
        # #     print(f"\t\t{c.compatibility}")
        # logger.info(f"\tNumber of mappings {for_einsum_text}: {nmappings}")
        # logger.info(
        #     f"\tMappings per group {for_einsum_text}: {nmappings / len(combined)}"
        # )
        # logger.info(
        #     f"\tLargest left: {max(len(s2.mappings.data) for s in left.values() for s2, _ in s)}"
        # )
        # logger.info(
        #     f"\tLargest right: {max(len(s2.mappings.data) for s in right.values() for s2, _ in s)}"
        # )

        # ======================================================================
        # Update left for the next iteration.
        # =================================================================
        left = combined
        left_einsum = right_einsum
        left_tensors |= right_tensors

    # ======================================================================
    # Final consolidate and group
    # ======================================================================
    t0 = time.time()
    left = PmappingGroup.left_consolidate(
        left, None, pbar="Final consolidate" if print_progress else None
    )
    s_final = PmappingGroup.combine_combineable(
        left, oset(), print_progress=print_progress
    )
    assert len(s_final) == 1
    mappings = s_final[0].mappings
    mappings.limit_capacity(next_shared_loop_index=-1, finished=True)
    mappings.free_to_loop_index(-2)
    mappings.make_pareto()

    timer.log_total_time()
    # if evaluations_tracker is not None and "Total_latency" in data.columns and "Total_energy" in data.columns:
    #     edp = data["Total_latency"] * data["Total_energy"]
    #     edp_min = edp.min()
    #     evaluations_tracker.add_evaluation(n_evaluations, edp_min)
    #     evaluations_tracker.n_mappings.update(n_mappings)
    #     evaluations_tracker.runtime.update(runtime)

    return mappings


def _check_einsum2pmappings_not_empty(
    einsum2pmappings: dict[EinsumName, list[PmappingGroup]],
    pmappings: MultiEinsumPmappings,
):
    for einsum_name, einsum_pmappings in einsum2pmappings.items():
        total = sum(len(p.mappings.data) for p in einsum_pmappings)
        n_compatibilities = len(einsum_pmappings)
        logger.info(
            f"Einsum {einsum_name} has {total} pmappings with {n_compatibilities} compatibilities"
        )
        if total == 0:
            if einsum_name in pmappings.einsums_with_pmappings_generated:
                keep_rates = pmappings.pmapping_keep_rates(per_einsum=True)[einsum_name]
                keep_rates_text = "\n\t".join(
                    f"{k}: {v:.2e}" for k, v in keep_rates.items()
                )
                raise ValueError(
                    f"Einsum {einsum_name} has no pmappings. This likely means that "
                    f"no pmappings satisfied constraints for the Einsum. Please check "
                    f"the stats outputs from the MultiEinsumPmappings object returned "
                    f"by `af.mapper.FFM.make_pmappings(spec)`. The following are the "
                    f"keep rates (porportion of pmappings that are NOT pruned) for "
                    f"various causes of pmapping removal:\n\t{keep_rates_text}"
                )

            raise ValueError(
                f"Einsum {einsum_name} has no pmappings generated. It looks like you "
                "may have used `make_pmappings` with `einsum_names` set. You may set "
                "`require_all_einsums=False` to ignore this error and map only the "
                "Einsums that have pmappings."
            )


class MappingFromRow:
    def __init__(
        self,
        row: pd.Series,
        rank_variable_bounds: dict[str, dict[str, int]],
        einsum_names: list[EinsumName] | None = None,
    ):
        self.row = row
        self.rank_variable_bounds = rank_variable_bounds
        self.einsum_names = einsum_names

    def __call__(self, _for_model: bool = False) -> Mapping:
        return Mapping._from_pmappings(
            row2pmappings(self.row, self.einsum_names, self.rank_variable_bounds),
            rank_variable_bounds=self.rank_variable_bounds,
            _for_model=_for_model,
        )

    def _repr_svg_(self) -> str:
        return self.render()

    def render(self, **kwargs) -> str:
        return self().render(**kwargs)



import math
import islpy as isl

from accelforge.frontend.renames import RankVariable, Rank
from accelforge.frontend.workload import Workload, TensorName, Einsum, EinsumName


def get_einsum_operation_space(workload: Workload, einsum_name: str) -> isl.Set:
    """Return isl.Set of all operations in an einsum."""
    einsum_shape = workload.get_iteration_space_shape_isl_string(einsum_name)
    rank_variable_names = ",".join(
        map(str, workload.einsums[einsum_name].rank_variables)
    )
    try:
        return isl.Set(
            f"{{ {einsum_name}_operation[{rank_variable_names}] : {einsum_shape} }}"
        )
    except:
        raise Exception(f"Error creating isl.Set for {einsum_name}: {einsum_shape}")


def get_dim_bounds(isl_set: isl.Set) -> list[int]:
    bounds = []
    for i in range(isl_set.dim(isl.dim_type.set)):
        max_val = isl_set.dim_max_val(i)
        min_val = isl_set.dim_min_val(i)
        shape = max_val - min_val + 1  # max is inclusive
        try:
            bounds.append(shape.to_python())
        except:
            raise Exception(
                f"Shape is not an integer. Are all rank variables bounded? "
                f"Shape {shape} for rank variable {i} in {isl_set}"
            )
    return bounds


def get_rank_variable_bounds(
    workload: Workload, einsum_name: EinsumName
) -> dict[RankVariable, int]:
    """Return dictionary mapping rank variable name to bound."""
    operation_space = get_einsum_operation_space(workload, einsum_name)
    dim_shapes = get_dim_bounds(operation_space)
    return {
        rank_var: shape
        for rank_var, shape in zip(
            workload.einsums[einsum_name].rank_variables, dim_shapes
        )
    }


def get_projection_multi_aff(einsum: Einsum, tensor: TensorName) -> isl.MultiAff:
    """Return isl.MultiAff of projection from einsum to tensor."""
    rank_variables = einsum.rank_variables
    projection = einsum.tensor_accesses[tensor].projection

    rank_variables_str = ",".join(map(str, rank_variables))

    projection_str = ", ".join(
        f"{rank_name}={rank_projection}"
        for rank_name, rank_projection in projection.items()
    )

    return isl.MultiAff(
        f"{{ {einsum.name}_operation[{rank_variables_str}] -> "
        f"{tensor}[{projection_str}] }}"
    )


def get_projection_map(einsum: Einsum, tensor: TensorName) -> isl.Map:
    """Return isl.Map of projection from einsum to tensor."""
    return get_projection_multi_aff(einsum, tensor).as_map()


def get_tensor_data_space(workload: Workload, tensor: TensorName) -> isl.Set:
    """
    Get tensor data space based on the operation spaces of (for lack of
    a better term)'canonical' Einsums.

    Canonical Einsums (for this purpose) are all reader Einsums if the
    tensor is only ever read or all writer EInsums if the tensor is ever
    an output tensor.
    """
    writer_einsums = workload.einsums_with_tensor_as_output(tensor)
    if len(writer_einsums) == 0:
        reader_einsums = workload.einsums_with_tensor_as_input(tensor)
        canonical_einsums = reader_einsums
    else:
        canonical_einsums = writer_einsums

    tensor_data_space = None
    for einsum in canonical_einsums:
        operation_space = get_einsum_operation_space(workload, einsum.name)
        projection_map = get_projection_map(einsum, tensor)
        if tensor_data_space is None:
            tensor_data_space = operation_space.apply(projection_map)
        else:
            tensor_data_space = tensor_data_space.intersect(
                operation_space.apply(projection_map)
            )

    return tensor_data_space


def get_tensor_shape(workload: Workload, tensor: TensorName) -> dict[Rank, int]:
    bounds = get_dim_bounds(get_tensor_data_space(workload, tensor))
    an_access = next(iter(workload.accesses_for_tensor(tensor)))
    return {rank: bound for rank, bound in zip(an_access.ranks, bounds)}


def _card_box(data_space: isl.Set) -> int:
    dims = []
    for i in range(data_space.dim(isl.dim_type.set)):
        dim_min = data_space.dim_min(i)
        dim_max = data_space.dim_max(i)

        if dim_min.is_cst() and dim_max.is_cst():
            min_val = dim_min.as_aff().get_constant_val().to_python()
            max_val = dim_max.as_aff().get_constant_val().to_python()
        else:
            raise ValueError(f"Data space is not rectangular: {data_space}")

        dims.append(max_val - min_val + 1)

    return math.prod(dims)


ERRMSG = """ Non-box-shaped sets are not supported. This happens if ISL is installed without
Barvinok support. Please install ISL with Barvinok support, or use workloads with
rectangular data spaces and operation spaces. Non-rectangular spaces occur when the
workload contains complex expressions involving the rank variables, such as
multi-rank-variable inequalities.
Offending space: {space}.
"""


def get_tensor_size(workload: Workload, tensor: TensorName):
    """Get the size (num. of elements) of a tensor."""
    data_space = get_tensor_data_space(workload, tensor)
    if data_space.is_box():
        return _card_box(data_space)
    if not hasattr(data_space, "card"):
        raise RuntimeError(ERRMSG.format(space=str(data_space)))
    card_pwqp = data_space.card()
    return card_pwqp.eval(card_pwqp.domain().sample_point()).to_python()


def get_operation_space_size(workload: Workload, einsum_name: str):
    operation_space = get_einsum_operation_space(workload, einsum_name)
    if operation_space.is_box():
        return _card_box(operation_space)
    if not hasattr(operation_space, "card"):
        raise RuntimeError(ERRMSG.format(space=str(operation_space)))
    card_pwqp = operation_space.card()
    return card_pwqp.eval(card_pwqp.domain().sample_point()).to_python()



"""
Tests for workload model classes:
  - TensorAccess (properties: rank_variables, ranks, rank2rank_variables, etc.)
  - ImpliedProjection
  - Shape
  - Einsum (properties, validation, tensor_accesses)
  - Workload (validation, properties, einsums_with_tensor, etc.)
"""

import unittest

from accelforge.frontend.workload import (
    TensorAccess,
    ImpliedProjection,
    Shape,
    Einsum,
    Workload,
)

# ============================================================================
# TensorAccess
# ============================================================================


class TestTensorAccess(unittest.TestCase):
    """Tests for TensorAccess model and its properties."""

    def _make_access(self, name="T", projection=None, output=False, **kwargs):
        if projection is None:
            projection = ["m", "n"]
        return TensorAccess(name=name, projection=projection, output=output, **kwargs)

    def test_list_projection_becomes_implied(self):
        ta = self._make_access(projection=["a", "b", "c"])
        self.assertIsInstance(ta.projection, ImpliedProjection)
        self.assertEqual(ta.projection, {"A": "a", "B": "b", "C": "c"})

    def test_dict_projection_kept(self):
        ta = self._make_access(projection={"M": "m", "N": "n"})
        self.assertNotIsInstance(ta.projection, ImpliedProjection)
        self.assertEqual(ta.projection, {"M": "m", "N": "n"})

    def test_ranks_property(self):
        ta = self._make_access(projection=["a", "b", "c"])
        self.assertEqual(ta.ranks, ("A", "B", "C"))

    def test_ranks_property_dict(self):
        ta = self._make_access(projection={"X": "x", "Y": "y"})
        self.assertEqual(ta.ranks, ("X", "Y"))

    def test_rank_variables_property(self):
        ta = self._make_access(projection=["m", "n", "k"])
        self.assertEqual(ta.rank_variables, {"m", "n", "k"})

    def test_rank_variables_with_expression(self):
        ta = self._make_access(projection={"M": "m+n", "K": "k"})
        self.assertEqual(ta.rank_variables, {"m", "n", "k"})

    def test_directly_indexing_rank_variables(self):
        ta = self._make_access(projection={"M": "m", "N": "n"})
        self.assertEqual(ta.directly_indexing_rank_variables, {"m", "n"})

    def test_directly_indexing_simple_only(self):
        """Only simple single-identifier projections count as directly indexing."""
        ta = self._make_access(projection={"M": "m", "N": "n"})
        # Both are simple direct indices
        self.assertEqual(ta.directly_indexing_rank_variables, {"m", "n"})
        self.assertEqual(ta.expression_indexing_rank_variables, set())

    def test_expression_indexing_rank_variables(self):
        """Rank variables that appear only in expressions (not as standalone indices)."""
        ta = self._make_access(projection={"M": "m", "K": "k"})
        self.assertEqual(ta.expression_indexing_rank_variables, set())

    def test_expression_indexing_with_complex_projection(self):
        """Expression-based projections have rank variables extracted from expressions."""
        ta = self._make_access(projection={"M": "m", "N": "n"})
        # All are direct, none are expression-based
        all_rv = ta.rank_variables
        direct = ta.directly_indexing_rank_variables
        expr = ta.expression_indexing_rank_variables
        self.assertEqual(expr, all_rv - direct)

    def test_rank2rank_variables(self):
        ta = self._make_access(projection={"M": "m", "N": "n"})
        result = ta.rank2rank_variables
        self.assertEqual(result["M"], {"m"})
        self.assertEqual(result["N"], {"n"})

    def test_rank_variable2ranks(self):
        ta = self._make_access(projection={"M": "m", "N": "m"})
        result = ta.rank_variable2ranks
        self.assertEqual(result["m"], {"M", "N"})

    def test_default_output_false(self):
        ta = self._make_access()
        self.assertFalse(ta.output)

    def test_output_true(self):
        ta = self._make_access(output=True)
        self.assertTrue(ta.output)

    def test_default_persistent_false(self):
        ta = self._make_access()
        self.assertFalse(ta.persistent)

    def test_bits_per_value_default_none(self):
        ta = self._make_access()
        self.assertIsNone(ta.bits_per_value)

    def test_bits_per_value_set(self):
        ta = self._make_access(bits_per_value=16)
        self.assertEqual(ta.bits_per_value, 16)

    def test_backing_storage_size_scale_default(self):
        ta = self._make_access()
        self.assertEqual(ta.backing_storage_size_scale, 1.0)

    def test_to_formatted_string(self):
        ta = self._make_access(name="W", projection=["k", "n"])
        s = ta._to_formatted_string()
        self.assertIn("W", s)

    def test_to_formatted_string_with_dict_projection(self):
        ta = self._make_access(name="K", projection={"B": "b", "M": "p"})
        s = ta._to_formatted_string()
        self.assertIn("K", s)


# ============================================================================
# ImpliedProjection
# ============================================================================


class TestImpliedProjection(unittest.TestCase):
    """Tests for ImpliedProjection subclass of dict."""

    def test_is_dict(self):
        ip = ImpliedProjection({"A": "a", "B": "b"})
        self.assertIsInstance(ip, dict)
        self.assertIsInstance(ip, ImpliedProjection)

    def test_values(self):
        ip = ImpliedProjection({"M": "m", "N": "n"})
        self.assertEqual(ip["M"], "m")
        self.assertEqual(ip["N"], "n")

    def test_empty(self):
        ip = ImpliedProjection()
        self.assertEqual(len(ip), 0)


# ============================================================================
# Shape
# ============================================================================


class TestShape(unittest.TestCase):
    """Tests for the Shape class (list of ISL expressions)."""

    def test_rank_variables_from_expressions(self):
        s = Shape(["0 <= m < 10", "0 <= n < 20"])
        self.assertEqual(s.rank_variables, {"m", "n"})

    def test_rank_variables_complex(self):
        s = Shape(["0 <= a + b < 10"])
        self.assertEqual(s.rank_variables, {"a", "b"})

    def test_empty_shape(self):
        s = Shape()
        self.assertEqual(s.rank_variables, set())

    def test_single_expression(self):
        s = Shape(["0 <= x < 5"])
        self.assertEqual(s.rank_variables, {"x"})


# ============================================================================
# Einsum
# ============================================================================


class TestEinsum(unittest.TestCase):
    """Tests for the Einsum model."""

    def _make_einsum(self, name="Matmul", **kwargs):
        defaults = {
            "tensor_accesses": [
                {"name": "A", "projection": ["m", "k"]},
                {"name": "B", "projection": ["k", "n"]},
                {"name": "C", "projection": ["m", "n"], "output": True},
            ],
        }
        defaults.update(kwargs)
        return Einsum(name=name, **defaults)

    def test_basic_creation(self):
        e = self._make_einsum()
        self.assertEqual(e.name, "Matmul")
        self.assertEqual(len(e.tensor_accesses), 3)

    def test_rank_variables(self):
        e = self._make_einsum()
        self.assertEqual(e.rank_variables, {"m", "k", "n"})

    def test_ranks(self):
        e = self._make_einsum()
        self.assertEqual(e.ranks, {"M", "K", "N"})

    def test_input_tensor_names(self):
        e = self._make_einsum()
        self.assertEqual(e.input_tensor_names, {"A", "B"})

    def test_output_tensor_names(self):
        e = self._make_einsum()
        self.assertEqual(e.output_tensor_names, {"C"})

    def test_tensor_names(self):
        e = self._make_einsum()
        self.assertEqual(e.tensor_names, {"A", "B", "C"})

    def test_tensor2rank_variables(self):
        e = self._make_einsum()
        t2rv = e.tensor2rank_variables
        self.assertEqual(t2rv["A"], {"m", "k"})
        self.assertEqual(t2rv["B"], {"k", "n"})
        self.assertEqual(t2rv["C"], {"m", "n"})

    def test_tensor2directly_indexing_rank_variables(self):
        e = self._make_einsum()
        t2di = e.tensor2directly_indexing_rank_variables
        self.assertEqual(t2di["A"], {"m", "k"})

    def test_tensor2irrelevant_rank_variables(self):
        e = self._make_einsum()
        t2irr = e.tensor2irrelevant_rank_variables
        self.assertEqual(t2irr["A"], {"n"})
        self.assertEqual(t2irr["B"], {"m"})
        self.assertEqual(t2irr["C"], {"k"})

    def test_rank_variable2ranks(self):
        e = self._make_einsum()
        rv2r = e.rank_variable2ranks
        self.assertEqual(rv2r["m"], {"M"})
        self.assertEqual(rv2r["k"], {"K"})
        self.assertEqual(rv2r["n"], {"N"})

    def test_indexing_expressions(self):
        e = self._make_einsum()
        self.assertEqual(e.indexing_expressions, {"m", "k", "n"})

    def test_reserved_name_total_raises(self):
        with self.assertRaises(ValueError) as ctx:
            self._make_einsum(name="Total")
        self.assertIn("reserved", str(ctx.exception).lower())

    def test_n_instances_default(self):
        e = self._make_einsum()
        self.assertEqual(e.n_instances, 1)

    def test_n_instances_custom(self):
        e = self._make_einsum(n_instances=5)
        self.assertEqual(e.n_instances, 5)

    def test_is_copy_operation_default_false(self):
        e = self._make_einsum()
        self.assertFalse(e.is_copy_operation)

    def test_copy_source_tensor_none_if_not_copy(self):
        e = self._make_einsum()
        self.assertIsNone(e.copy_source_tensor())

    def test_copy_source_tensor(self):
        e = Einsum(
            name="Copy",
            tensor_accesses=[
                {"name": "Src", "projection": ["m"]},
                {"name": "Dst", "projection": ["m"], "output": True},
            ],
            is_copy_operation=True,
        )
        self.assertEqual(e.copy_source_tensor(), "Src")

    def test_copy_with_multiple_inputs_raises(self):
        e = Einsum(
            name="Copy",
            tensor_accesses=[
                {"name": "Src1", "projection": ["m"]},
                {"name": "Src2", "projection": ["m"]},
                {"name": "Dst", "projection": ["m"], "output": True},
            ],
            is_copy_operation=True,
        )
        with self.assertRaises(ValueError):
            e.copy_source_tensor()

    def test_renames_as_dict(self):
        e = self._make_einsum(renames={"input": "A", "output": "C"})
        self.assertEqual(len(e.renames), 2)

    def test_renames_as_list(self):
        e = self._make_einsum(renames=[])
        self.assertEqual(len(e.renames), 0)

    def test_empty_tensor_accesses_rank_variables(self):
        e = Einsum(name="Empty", tensor_accesses=[])
        self.assertEqual(e.rank_variables, set())
        self.assertEqual(e.ranks, set())

    def test_to_formatted_string(self):
        e = self._make_einsum()
        s = e._to_formatted_string()
        self.assertIn("C", s)
        self.assertIn("A", s)


# ============================================================================
# Workload
# ============================================================================


class TestWorkload(unittest.TestCase):
    """Tests for the Workload model."""

    def _make_workload(self, **kwargs):
        defaults = {
            "rank_sizes": {"M": 64, "K": 32, "N": 64},
            "bits_per_value": {"All": 8},
            "einsums": [
                {
                    "name": "Matmul",
                    "tensor_accesses": [
                        {"name": "A", "projection": ["m", "k"]},
                        {"name": "B", "projection": ["k", "n"]},
                        {"name": "C", "projection": ["m", "n"], "output": True},
                    ],
                }
            ],
        }
        defaults.update(kwargs)
        return Workload(**defaults)

    def test_basic_creation(self):
        w = self._make_workload()
        self.assertEqual(len(w.einsums), 1)
        self.assertEqual(w.einsums[0].name, "Matmul")

    def test_einsum_names(self):
        w = self._make_workload()
        self.assertEqual(w.einsum_names, ["Matmul"])

    def test_tensor_names(self):
        w = self._make_workload()
        self.assertEqual(w.tensor_names, {"A", "B", "C"})

    def test_rank_variables(self):
        w = self._make_workload()
        self.assertEqual(w.rank_variables, {"m", "k", "n"})

    def test_einsums_with_tensor(self):
        w = self._make_workload()
        self.assertEqual(len(w.einsums_with_tensor("A")), 1)
        self.assertEqual(w.einsums_with_tensor("A")[0].name, "Matmul")
        self.assertEqual(len(w.einsums_with_tensor("NotHere")), 0)

    def test_einsums_with_tensor_as_input(self):
        w = self._make_workload()
        self.assertEqual(len(w.einsums_with_tensor_as_input("A")), 1)
        self.assertEqual(len(w.einsums_with_tensor_as_input("C")), 0)

    def test_einsums_with_tensor_as_output(self):
        w = self._make_workload()
        self.assertEqual(len(w.einsums_with_tensor_as_output("C")), 1)
        self.assertEqual(len(w.einsums_with_tensor_as_output("A")), 0)

    def test_accesses_for_tensor(self):
        w = self._make_workload()
        accesses = w.accesses_for_tensor("A")
        self.assertEqual(len(accesses), 1)
        self.assertEqual(accesses[0].name, "A")

    def test_n_instances_default(self):
        w = self._make_workload()
        self.assertEqual(w.n_instances, 1)

    def test_duplicate_einsum_names_raises(self):
        with self.assertRaises(ValueError) as ctx:
            Workload(
                rank_sizes={"B": 4, "M": 8},
                bits_per_value={"All": 8},
                einsums=[
                    "V[b, m] = I[b]",
                    "V[b, m] = X[b]",
                ],
            )
        self.assertIn("not unique", str(ctx.exception))

    def test_inconsistent_tensor_ranks_raises(self):
        with self.assertRaises(ValueError) as ctx:
            Workload(
                rank_sizes={"B": 4, "M": 8, "D": 16},
                bits_per_value={"All": 8},
                einsums=[
                    "V[b, m] = I[b]",
                    "K[b, m, d] = I[b, d]",
                ],
            )
        self.assertIn("inconsistent ranks", str(ctx.exception))

    def test_concise_string_einsums(self):
        w = Workload(
            rank_sizes={"B": 4, "M": 8, "D": 16},
            bits_per_value={"All": 8},
            einsums=[
                "V[b, m] = I[b, m] * W[m]",
            ],
        )
        self.assertEqual(len(w.einsums), 1)
        self.assertEqual(w.einsums[0].name, "V")

    def test_multi_einsum_workload(self):
        w = Workload(
            rank_sizes={"M": 64, "N0": 32, "N1": 32, "N2": 32},
            bits_per_value={"All": 8},
            einsums=[
                {
                    "name": "Matmul1",
                    "tensor_accesses": [
                        {"name": "T0", "projection": ["m", "n0"]},
                        {"name": "W0", "projection": ["n0", "n1"]},
                        {"name": "T1", "projection": ["m", "n1"], "output": True},
                    ],
                },
                {
                    "name": "Matmul2",
                    "tensor_accesses": [
                        {"name": "T1", "projection": ["m", "n1"]},
                        {"name": "W1", "projection": ["n1", "n2"]},
                        {"name": "T2", "projection": ["m", "n2"], "output": True},
                    ],
                },
            ],
        )
        self.assertEqual(len(w.einsums), 2)
        self.assertEqual(w.tensor_names_used_in_multiple_einsums, {"T1"})

    def test_tensor_copies(self):
        w = Workload(
            rank_sizes={"M": 64},
            bits_per_value={"All": 8},
            einsums=[
                {
                    "name": "Copy",
                    "tensor_accesses": [
                        {"name": "Src", "projection": ["m"]},
                        {"name": "Dst", "projection": ["m"], "output": True},
                    ],
                    "is_copy_operation": True,
                },
            ],
        )
        copies = w.get_tensor_copies()
        self.assertIn("Src", copies)
        self.assertIn("Dst", copies["Src"])

    def test_empty_workload(self):
        w = Workload()
        self.assertEqual(len(w.einsums), 0)
        self.assertEqual(w.tensor_names, set())
        self.assertEqual(w.rank_variables, set())

    def test_iteration_space_shape(self):
        w = Workload(
            iteration_space_shape={"m": "0 <= m < 64", "n": "0 <= n < 32"},
            rank_sizes={"M": 64, "N": 32},
            bits_per_value={"All": 8},
            einsums=[
                {
                    "name": "E",
                    "tensor_accesses": [
                        {"name": "A", "projection": ["m"]},
                        {"name": "B", "projection": ["m", "n"], "output": True},
                    ],
                },
            ],
        )
        self.assertIn("m", w.iteration_space_shape)

    def test_persistent_tensors_field(self):
        w = Workload(
            rank_sizes={"M": 64},
            bits_per_value={"All": 8},
            persistent_tensors="weight - Intermediates",
            einsums=[],
        )
        self.assertEqual(w.persistent_tensors, "weight - Intermediates")


if __name__ == "__main__":
    unittest.main()



from typing import Generic, TypeVar


T = TypeVar("T")


def _sorted_iter(iterable):
    """Sort elements for deterministic iteration. Falls back to str() keys
    for elements that are not natively comparable."""
    items = list(iterable)
    if not items:
        return iter(items)
    try:
        return iter(sorted(items))
    except TypeError:
        return iter(sorted(items, key=lambda x: str(x)))


class fzs(frozenset[T], Generic[T]):
    def __repr__(self):
        return f"{{{', '.join(sorted(x.__repr__() for x in self))}}}"

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        return _sorted_iter(frozenset.__iter__(self))

    def __or__(self, other: "fzs[T]") -> "fzs[T]":
        return fzs(super().__or__(other))

    def __and__(self, other: "fzs[T]") -> "fzs[T]":
        return fzs(super().__and__(other))

    def __sub__(self, other: "fzs[T]") -> "fzs[T]":
        return fzs(super().__sub__(other))

    def __xor__(self, other: "fzs[T]") -> "fzs[T]":
        return fzs(super().__xor__(other))

    def __lt__(self, other: "fzs[T]") -> bool:
        return sorted(self) < sorted(other)

    def __le__(self, other: "fzs[T]") -> bool:
        return sorted(self) <= sorted(other)

    def __gt__(self, other: "fzs[T]") -> bool:
        return sorted(self) > sorted(other)

    def __ge__(self, other: "fzs[T]") -> bool:
        return sorted(self) >= sorted(other)


class oset(set, Generic[T]):
    """Set that iterates in sorted order for deterministic behavior."""

    def __repr__(self):
        items = ", ".join(repr(x) for x in self)
        return f"oset({{{items}}})"

    def __iter__(self):
        return _sorted_iter(set.__iter__(self))

    def pop(self):
        val = min(self)
        self.discard(val)
        return val

    def __or__(self, other):
        return oset(set.__or__(self, other))

    def __ror__(self, other):
        return oset(set.__ror__(self, other))

    def __and__(self, other):
        return oset(set.__and__(self, other))

    def __rand__(self, other):
        return oset(set.__rand__(self, other))

    def __sub__(self, other):
        return oset(set.__sub__(self, other))

    def __rsub__(self, other):
        return oset(set.__rsub__(self, other))

    def __xor__(self, other):
        return oset(set.__xor__(self, other))

    def __rxor__(self, other):
        return oset(set.__rxor__(self, other))

    def copy(self):
        return oset(set.copy(self))

    def union(self, *others):
        return oset(set.union(self, *others))

    def intersection(self, *others):
        return oset(set.intersection(self, *others))

    def difference(self, *others):
        return oset(set.difference(self, *others))

    def symmetric_difference(self, other):
        return oset(set.symmetric_difference(self, other))



import functools
from math import ceil, comb
import functools
import math
import numbers
import pandas as pd
import numpy as np
from accelforge.util._frozenset import oset

NUMPY_FLOAT_TYPE = np.float32


@functools.lru_cache(maxsize=None)
def _count_factorizations_imperfect(n, into_n_parts):
    # Factorize n into into_n_parts parts
    RUBY_STYLE_IMPERFECT = True
    # RUBY_STYLE_IMPERFECT = True
    if n <= 1:
        return 1
    if into_n_parts <= 0:
        return 1

    shapes = list(range(1, ceil(n**0.5) + 1))
    shapes = shapes + [ceil(n / s) for s in shapes]
    shapes = sorted(oset(shapes))

    if RUBY_STYLE_IMPERFECT:
        shapes = list(range(1, n + 1))

    total = 0
    for s in shapes:
        n = _count_factorizations_imperfect(ceil(n / s), into_n_parts - 1)
        total += _count_factorizations_imperfect(ceil(n / s), into_n_parts - 1)

    return total


def _prime_factorization(n):
    f = []
    i = 2
    while n > 1:
        if n % i == 0:
            f.append(i)
            n //= i
        else:
            i += 1
    return f


def _count_factorizations(n, into_n_parts, imperfect=False):
    if into_n_parts <= 1:
        return 1
    f = _prime_factorization(n)
    factors = {f2: f.count(f2) for f2 in oset(f)}
    total = 1
    for exp in factors.values():
        total *= comb(exp + into_n_parts - 1, into_n_parts - 1)  # n choose k

    if imperfect:
        n = _count_factorizations_imperfect(n, into_n_parts)
        assert n >= total, f"n: {n} < total: {total}"
        return n

    return total


def _fillna_and__numeric_cast(df: pd.DataFrame, value: float) -> pd.DataFrame:
    def _is_float(x) -> bool:
        return isinstance(x, numbers.Real)

    def _is_int(x) -> bool:
        return (
            isinstance(x, numbers.Integral)
            or isinstance(x, numbers.Real)
            and (math.isnan(x) or int(x) == x)
        )

    for col in [c for c in df.columns if df.dtypes[c] == object]:
        # If it's an object col and all of them are integers, convert to int. nans count
        # as True
        if all(_is_int(x) for x in df[col]):
            df[col] = df[col].replace(float("nan"), value).astype(int)
        elif all(_is_float(x) for x in df[col]):
            df[col] = df[col].replace(float("nan"), value).astype(float)

    cols = df.select_dtypes(include=[np.floating, float, np.integer, int]).columns
    df[cols] = df[cols].fillna(value)
    for col in df.columns:
        assert (
            not df[col].isna().any()
        ), f"df has nans in column {col} with dtype {df[col].dtype}. " + "\n".join(
            f"{x} {type(x)=} {_is_int(x)=} {_is_float(x)=} " for x in df[col]
        )
    return df


def _numeric_cast(df: pd.DataFrame) -> pd.DataFrame:
    def _is_float(x) -> bool:
        return isinstance(x, numbers.Real)

    def _is_int(x) -> bool:
        return (
            isinstance(x, numbers.Integral)
            or isinstance(x, numbers.Real)
            and int(x) == x
        )

    for col in [c for c in df.columns if df.dtypes[c] == object]:
        series = df[col]
        if all(_is_int(x) for x in series):
            df[col] = series.astype(int)
        elif all(_is_float(x) for x in series):
            df[col] = series.astype(NUMPY_FLOAT_TYPE)
    return df



import functools
from math import prod
import time

from accelforge._accelerated_imports import pandas as pd

from paretoset import paretoset
from joblib import delayed
from sympy import factorint

from accelforge._accelerated_imports import np
from accelforge.util._frozenset import oset
from accelforge.util.parallel import parallel

from accelforge.mapper.FFM._pareto_df.df_convention import (
    col2reservation,
    col_used_in_pareto,
    is_fused_loop_col,
    is_n_iterations_col,
    is_objective_col,
)

from .fast_pareto import fast_pareto_mask


def dominates(a: pd.Series, b: pd.Series) -> bool:
    return all(a[i] <= b[i] for i in range(len(a)))


def check_dominance(df: pd.DataFrame, n_optimal: int):
    # mask = np.zeros(len(df), dtype=bool)
    # mask[:new_point] = True
    mask = np.zeros(len(df) - n_optimal, dtype=bool)
    for col in df.columns:
        compare = df.iloc[n_optimal - 1][col]
        mask = mask | (df[col].iloc[n_optimal:] < compare)
    return np.concatenate([np.ones(n_optimal, dtype=bool), mask])


def quickpareto(df: pd.DataFrame) -> pd.DataFrame:
    # Step 1: Sort by the column with the most unique values
    # Step 2: Extract the first row. Add it to the pareto set
    # Step 3: Remove all dominated points
    # Step 4: Repeat until no more points to add

    # Step 1: Sort by the column with the most unique values
    original_len = len(df)
    col_to_sort = max(df.columns, key=lambda c: df[c].nunique())
    df = df.sort_values(by=col_to_sort).drop(columns=[col_to_sort])

    new_point = 0
    while new_point < len(df):
        mask = check_dominance(df, new_point + 1)
        df = df[mask]
        new_point += 1

    # Turn the index into a mask
    mask = np.zeros(original_len, dtype=bool)
    mask[df.index] = True
    return mask


def makepareto_quick2(mappings: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    from fast_pareto import is_pareto_front

    m2 = mappings[columns]
    m2 = m2[is_pareto_front(m2.to_numpy())].drop_duplicates()
    return mappings.loc[m2.index]


def makepareto_quick(mappings: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    return mappings[quickpareto(mappings[columns])]


def paretofy_chunk(chunk, sense: list[str]):
    return fast_pareto_mask(chunk.values if hasattr(chunk, "values") else chunk, sense)


def makepareto_merge(
    mappings: pd.DataFrame,
    columns: list[str],
    parallelize: bool = False,
    split_by_cols: list[str] = (),
) -> pd.DataFrame:
    if len(mappings) <= 1:
        return mappings

    sense = ["min"] * len(columns) + ["diff"] * len(split_by_cols)
    combined = mappings[columns + list(split_by_cols)]
    return mappings[fast_pareto_mask(combined.values, sense)]


def makepareto_time_compare(mappings: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    t0 = time.time()
    pareto = makepareto_merge(mappings, columns)
    t1 = time.time()
    merge_time = t1 - t0
    print(
        f"Time to make pareto with merge: {t1 - t0: .2f}. Number of pareto points: {len(pareto)}"
    )

    t0 = time.time()
    pareto2 = makepareto_quick2(mappings, columns)
    t1 = time.time()
    print(
        f"Time to make pareto with quick: {t1 - t0: .2f}. Number of pareto points: {len(pareto2)}"
    )
    quick_time = t1 - t0

    print(f"Quick is {quick_time / merge_time: .2f}x slower")

    if len(pareto) != len(pareto2):
        print(f"mismatch: {len(pareto)} != {len(pareto2)}")
        makepareto_quick2(mappings)

    return pareto2


# 2d. Blockwise vectorized CuPy Pareto front with sorting by one objective (full check)
# 2c. Fully vectorized CuPy brute-force Pareto front
# (returns numpy mask for compatibility)
def pareto_front_cupy_vectorized(X):
    # if len(X) > 1000:
    #     return X[paretoset(X.get(), sense=["min"] * X.shape[1])]

    # Broadcast X_gpu to (n, n, m) for all-pairs comparison
    A = X[:, None, :]  # shape (n, 1, m)
    B = X[None, :, :]  # shape (1, n, m)
    less_equal = (B <= A).all(axis=2)  # shape (n, n)
    strictly_less = (B < A).any(axis=2)  # shape (n, n)
    dominated = less_equal & strictly_less  # shape (n, n)
    is_pareto = ~dominated.any(axis=1)
    return is_pareto


# 2d. Recursive blockwise merge CuPy Pareto front with sorting by one objective
def pareto_front_cupy_blockwise_sorted_recursive(X, block_size=2000):
    N = X.shape[0]
    if N <= block_size:
        # Base case: just compute Pareto front directly
        mask = pareto_front_cupy_vectorized(X)
        return mask
    # Split into two halves
    mid = N // 2
    a, b = X[:mid], X[mid:]
    mask_a = pareto_front_cupy_blockwise_sorted_recursive(a, block_size)
    mask_b = pareto_front_cupy_blockwise_sorted_recursive(b, block_size)
    # Get Pareto-optimal points from both halves
    pareto_points_a = a[mask_a]
    pareto_points_b = b[mask_b]
    merged_points = np.vstack([pareto_points_a, pareto_points_b])
    # Compute Pareto front of the merged set
    merged_mask = pareto_front_cupy_vectorized(merged_points)
    merged_indices = np.where(merged_mask)[0]
    # Map merged_indices back to the original indices in X
    # First, get the indices in X for the merged points
    indices_a = np.where(mask_a)[0]
    indices_b = np.where(mask_b)[0] + mid
    all_indices = np.concatenate([indices_a, indices_b])
    merged_indices_in_X = all_indices[merged_indices]
    # Build the final mask for X
    mask = np.zeros(N, dtype=bool)
    mask[merged_indices_in_X] = True
    return mask


# def makepareto(
#     mappings: pd.DataFrame,
#     columns: list[str] = None,
#     parallelize: bool = False,
#     split_by_cols: list[str] = (),
# ) -> pd.DataFrame:
#     # return makepareto_time_compare(mappings)
#     if columns is None:
#         columns = [c for c in mappings.columns if col_used_in_pareto(c)]
#     if _accelerated_imports.ACCELERATED:
#         mask = pareto_front_cupy_blockwise_sorted_recursive(mappings[columns].to_cupy())
#         return mappings[mask]


def round_to_tolerance(x: pd.Series, tolerance: float) -> pd.Series:
    if tolerance == 0:
        return x
    x = np.round(x / tolerance) * tolerance
    return x


def logscale_to_tolerance(x: pd.Series, tolerance: float) -> pd.Series:
    if tolerance == 0 or tolerance is None:
        return x
    assert tolerance > 0
    if x.min() <= 0:
        return x
    rounded = np.round(np.log(x) / np.log(1 + tolerance))
    return np.exp(rounded * np.log(1 + tolerance))


def multi_round(
    x: np.ndarray, tolerance: float, absolute_tolerance: float
) -> np.ndarray:
    tolerance = 0 if tolerance is None else tolerance
    absolute_tolerance = 0 if absolute_tolerance is None else absolute_tolerance
    if tolerance == 0:
        return round_to_tolerance(x, absolute_tolerance)
    if absolute_tolerance == 0:
        return logscale_to_tolerance(x, tolerance)

    abs_step_size = absolute_tolerance
    log_step_size = tolerance * x

    rounded = round_to_tolerance(x, abs_step_size)
    logscale_rounded = logscale_to_tolerance(x, tolerance)

    use_log_mask = log_step_size > abs_step_size
    new = np.where(use_log_mask, logscale_rounded, rounded)
    return new


def makepareto(
    mappings: pd.DataFrame,
    columns: list[str] = None,
    split_by_cols: list[str] = (),
    resource_usage_tolerance: float = 0,
    objective_tolerance: float = 0,
    absolute_resource_usage_tolerance: float = 0,
) -> pd.DataFrame:
    # return makepareto_time_compare(mappings)
    if columns is None:
        columns = [c for c in mappings.columns if col_used_in_pareto(c)]

    # Number of iterations is derived from the tile shapes, so we don't need to use it,
    # since any row with the same tile shapes will have the same number of iterations.
    split_by_cols = list(split_by_cols) + [
        c
        for c in mappings.columns
        if is_fused_loop_col(c) and not is_n_iterations_col(c)
    ]

    goals = []
    to_pareto = []
    for c in mappings.columns:
        if mappings[c].nunique() <= 1:
            continue

        if c in columns and is_objective_col(c):
            to_pareto.append(logscale_to_tolerance(mappings[c], objective_tolerance))
            goals.append("min")
        elif c in split_by_cols:
            to_pareto.append(mappings[c])
            goals.append("diff")
        elif c in columns:
            x = mappings[c]
            if col2reservation(c) is not None:
                x = multi_round(
                    x, resource_usage_tolerance, absolute_resource_usage_tolerance
                )
            to_pareto.append(x)
            goals.append("min")

    if not to_pareto:
        return mappings.iloc[0:1]

    # import os
    # import uuid
    # import pickle as pkl

    # os.makedirs("/tmp/paretos", exist_ok=True)
    # pkl.dump(
    #     (pd.concat(to_pareto, axis=1), goals),
    #     open(f"/tmp/paretos/{uuid.uuid4()}.pkl", "wb")
    # )

    combined = pd.concat(
        (pd.Series(p).reset_index(drop=True) for p in to_pareto), axis=1
    )
    return mappings[fast_pareto_mask(combined.values, goals)]


@functools.lru_cache(maxsize=10000)
def _factorint_cached(x: int):
    return factorint(x)


def prime_factor_counts(arr: np.ndarray) -> np.ndarray:
    arr = np.asarray(arr, dtype=int)
    unique_vals = np.unique(arr)
    factorizations = {x: _factorint_cached(x) for x in unique_vals}

    # Gather all unique primes
    all_primes = sorted(oset(p for f in factorizations.values() for p in f))

    # Build result matrix
    result = np.zeros((len(arr), len(all_primes)), dtype=int)
    prime_index = {p: j for j, p in enumerate(all_primes)}

    for i, x in enumerate(arr.flatten()):
        for p, exp in factorizations[x].items():
            result[i, prime_index[p]] = exp

    return result


def makepareto_numpy(
    mappings: np.ndarray,
    goals: list[str],
    dirty: bool = False,  # Doesn't do anything
    tolerances: list[float] = None,
    absolute_tolerances: list[float] = None,
) -> pd.DataFrame:

    to_pareto = []
    new_goals = []
    if tolerances is None:
        tolerances = [0] * len(goals)
    if absolute_tolerances is None:
        absolute_tolerances = [0] * len(goals)

    assert len(goals) == mappings.shape[1]
    for c in range(mappings.shape[1]):
        if len(np.unique(mappings[:, c])) <= 1:
            continue

        goal = goals[c]
        rounded = multi_round(mappings[:, c], tolerances[c], absolute_tolerances[c])

        if goal in ["min", "max"]:
            to_pareto.append(rounded if goal == "min" else -rounded)
            new_goals.append("min")
        elif goal == "diff":
            to_pareto.append(rounded)
            new_goals.append("diff")
        elif goal == "min_per_prime_factor":
            counts = prime_factor_counts(rounded)
            for i in range(counts.shape[1]):
                to_pareto.append(counts[:, i])
                new_goals.append("min")
        elif goal == "max_per_prime_factor":
            counts = prime_factor_counts(rounded)
            for i in range(counts.shape[1]):
                to_pareto.append(counts[:, i])
                new_goals.append("max")
        else:
            raise ValueError(f"Unknown goal: {goal}")

    if not to_pareto:
        n = np.zeros(mappings.shape[0], dtype=bool)
        n[0] = True
        return n

    data = np.concatenate([x.reshape(-1, 1) for x in to_pareto], axis=1)
    return fast_pareto_mask(data, new_goals)



import functools
from accelforge.util.exceptions import EvaluationError
from accelforge.util._frozenset import _sorted_iter, oset
from pydantic import BaseModel, ConfigDict, model_serializer
from typing import Iterator, Optional, TypeVar, Generic, Any, Union
from accelforge.util._eval_expressions import MATH_FUNCS

T = TypeVar("T")


def _reconstruct_invertible_set(state):
    """Helper function to reconstruct InvertibleSet during unpickling."""
    obj = object.__new__(InvertibleSet)
    obj.__dict__.update(state)
    return obj


class InvertibleSet(BaseModel, Generic[T]):
    instance: frozenset[T]
    full_space: frozenset[T]
    space_type: type[T]
    # child_access_name: Optional[str] = None
    element_to_child_space: Optional[dict[str, Any]] = None
    _bits_per_value: Optional[int] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @model_serializer
    def _serialize_model(self):
        """Custom serializer for InvertibleSet to avoid Pydantic serialization warnings."""
        return {
            "instance": list(self.instance),
            "full_space": list(self.full_space),
            "space_type": self.space_type.__name__,
            "element_to_child_space": self.element_to_child_space,
            "_bits_per_value": self._bits_per_value,
        }

    @property
    def bits_per_value(self) -> int:
        if len(self.instance) != 1:
            raise ValueError(
                f"Can not access bits_per_value for a set !=1 elements: "
                f"{self.instance}."
            )
        if self._bits_per_value is None:
            raise ValueError(f"Bits per value is not defined for set {self.instance}.")
        return self._bits_per_value

    @bits_per_value.setter
    def bits_per_value(self, value: int):
        self._bits_per_value = value

    def __reduce__(self):
        return (_reconstruct_invertible_set, (self.__dict__,))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __deepcopy__(self, memo):
        """Custom deepcopy implementation to avoid pydantic deepcopy issues."""
        import copy

        cls = type(self)
        # Create a new instance without calling __init__
        new_obj = cls.__new__(cls)
        # Mark it in the memo to handle circular references
        memo[id(self)] = new_obj
        # Deep copy the __dict__ directly to avoid triggering setattr
        new_obj.__dict__.update(copy.deepcopy(self.__dict__, memo))
        # Initialize pydantic's internal attributes if they don't exist
        if not hasattr(new_obj, "__pydantic_fields_set__"):
            object.__setattr__(new_obj, "__pydantic_fields_set__", set())
        if not hasattr(new_obj, "__pydantic_extra__"):
            object.__setattr__(new_obj, "__pydantic_extra__", {})
        if not hasattr(new_obj, "__pydantic_private__"):
            object.__setattr__(new_obj, "__pydantic_private__", {})
        return new_obj

    def __repr__(self):
        return f"InvertibleSet({self.instance})"

    def __str__(self):
        return self.__repr__()

    def __invert__(self):
        return self.to_my_space(self.full_space - self.instance)

    def check_match_space_name(self, other):
        if self.space_type != other.space_type:
            raise ValueError(
                f"Can not perform set operations between different spaces "
                f"{self.space_type} and {other.space_type}."
            )

    def to_my_space(self, other) -> Union[set, "InvertibleSet"]:
        return InvertibleSet(
            instance=other.instance if isinstance(other, InvertibleSet) else other,
            full_space=self.full_space,
            space_type=self.space_type,
            # child_access_name=self.child_access_name,
            element_to_child_space=self.element_to_child_space,
        )

    @staticmethod
    def _make_set(x) -> set:
        return x.instance if isinstance(x, InvertibleSet) else x

    def __and__(self, other: "InvertibleSet[T]") -> "InvertibleSet[T]":
        a, b = self._make_set(self), self._make_set(other)
        return self.to_my_space(a & b)

    def __or__(self, other: "InvertibleSet[T]") -> "InvertibleSet[T]":
        a, b = self._make_set(self), self._make_set(other)
        return self.to_my_space(a | b)

    def __sub__(self, other: "InvertibleSet[T]") -> "InvertibleSet[T]":
        a, b = self._make_set(self), self._make_set(other)
        return self.to_my_space(a - b)

    def __xor__(self, other: "InvertibleSet[T]") -> "InvertibleSet[T]":
        a, b = self._make_set(self), self._make_set(other)
        return self.to_my_space(a ^ b)

    def __call__(self):
        return self

    def _cast_to_child_space(self, *args, **kwargs):
        if not self.full_space:
            raise ValueError(f"Full space is empty for set {self.space_type}.")
        for item in self:
            if item not in self.element_to_child_space:
                raise ValueError(
                    f"Item {item} is not in the element_to_child_space "
                    f"for set {self.space_type}."
                )

        if not self.element_to_child_space:
            raise ValueError(
                f"Element to child space is not set for set {self.space_type}."
            )

        first_child_space_item: InvertibleSet = next(
            iter(self.element_to_child_space.values())
        )
        return first_child_space_item.to_my_space(
            oset.union(
                oset(), *(oset(self.element_to_child_space[item]) for item in self)
            )
        )

    def __bool__(self):
        return bool(self.instance)

    def __len__(self):
        return len(self.instance)

    def __contains__(self, item):
        return item in self.instance

    def __iter__(self):
        return _sorted_iter(self.instance)

    def __getitem__(self, item):
        return self.instance[item]

    def iter_one_element_sets(self) -> Iterator["InvertibleSet[T]"]:
        for item in _sorted_iter(self.instance):
            yield InvertibleSet(
                instance=oset((item,)),
                full_space=self.full_space,
                space_type=self.space_type,
                # child_access_name=self.child_access_name,
                element_to_child_space=self.element_to_child_space,
            )

    @property
    def rank_variables(self) -> set["RankVariable"]:
        from accelforge.frontend.workload import RankVariable
        from accelforge.frontend.renames import TensorName

        if self.space_type == TensorName:
            return self._cast_to_child_space()
        raise ValueError(
            f"Can not get rank variables for a set with space type "
            f"{self.space_type.__name__}."
        )

    @property
    def tensors(self) -> set["TensorName"]:
        from accelforge.frontend.renames import TensorName

        if self.space_type == TensorName:
            return self
        raise ValueError(
            f"Can not get tensors for a set with space type "
            f"{self.space_type.__name__}."
        )


def set_expression_type_check(
    result: InvertibleSet[T],
    expected_space: type[T],
    expected_count: int | None = None,
    location: str | None = None,
) -> None:
    if not isinstance(result, InvertibleSet):
        raise TypeError(f"Expected a InvertibleSet, got {type(result)}: {result}")
    if expected_space is not None and result.space_type != expected_space:
        raise ValueError(
            f"Expected a set with space type '{expected_space.__name__}', got {result.space_type.__name__}"
        )
    if expected_count is not None and len(result) != expected_count:
        raise ValueError(
            f"Expected {expected_count=} elements, got {len(result)}: {result.instance}"
        )


def eval_set_expression(
    expression: str | InvertibleSet,
    symbol_table: dict[str, InvertibleSet],
    expected_space: type[T],
    location: str,
    expected_count: int | None = None,
) -> InvertibleSet[T]:
    try:
        err = None
        if not isinstance(expression, (InvertibleSet, str)):
            raise TypeError(f"Expected a string, got {type(expression)}: {expression}")

        prev_result = "NOT_FOUND"
        if isinstance(expression, str):
            result = prev_result
            if expression in symbol_table:
                result = symbol_table[expression]
            elif expression[-2:] == "()" and expression[:-2] in symbol_table:
                try:
                    result = symbol_table[expression[:-2]]()
                except:
                    pass
        else:
            result = expression

        if id(result) == id(prev_result):
            result = eval(expression, {"__builtins__": MATH_FUNCS}, symbol_table)

        if not isinstance(result, InvertibleSet):
            raise TypeError(
                f"Returned a non-InvertibleSet with type {type(result)}: {result}"
            )
        set_expression_type_check(result, expected_space, expected_count, location)

    except Exception as e:

        def strformat(v):
            v = str(v)
            return v if len(v) <= 100 else v[:100] + "..."

        err = EvaluationError(
            f'{e}. Set expression: "{expression}". Symbol table:\n\t'
            + "\n\t".join(f"{k}: {strformat(v)}" for k, v in symbol_table.items())
        )
        if location is not None:
            err.add_field(location)
    if err:
        raise err
    return result



"""Applies the binding layer into one that can be used for later analysis,"""

from accelforge.frontend.binding import Binding
from accelforge.frontend.mapping import Mapping
from accelforge.frontend.workload import Workload

from accelforge.model._looptree.reuse.isl.mapping_to_isl.analyze_mapping import (
    occupancies_from_mapping,
    MappingAnalysisResult,
)


def apply_binding(binding: Binding, mapping: Mapping, workload: Workload):
    """
    Given a mapping, apply the mapping from logical components onto physical
    components.
    """
    map_analysis: MappingAnalysisResult = occupancies_from_mapping(mapping, workload)
    for buffet, occ in map_analysis.buffet_to_occupancy.items():
        print(occ)
    print(binding)



"""
Tests covering previously-untested API surface areas:

  1. Workload.get_iteration_space_shape_isl_string()
  2. Workload.n_instances / Einsum.n_instances
  3. Tensors.back, tile_shape, no_refetch_from_above, tensor_order_options
  4. Arch.total_area, total_leak_power, per_component_total_area/leak
  5. Spec.calculate_component_area_energy_latency_leak()
  6. Action.latency_scale, Component.latency_scale
  7. TensorHolder.bits_per_value_scale, bits_per_action
  8. Spec.to_yaml() round-trip serialization
  9. EvalableModel utility methods (all_fields_default, model_dump_non_none, etc.)
"""

import math
import os
import tempfile
import unittest
from pathlib import Path

from accelforge.frontend.spec import Spec
from accelforge.frontend.workload import (
    Workload,
    Einsum,
    TensorAccess,
    ImpliedProjection,
    Shape,
)
from accelforge.frontend.arch import (
    Action,
    Arch,
    Component,
    Comparison,
    Compute,
    Container,
    Memory,
    Tensors,
    TensorHolder,
    TensorHolderAction,
    Toll,
    Spatial as ArchSpatial,
)
from accelforge.frontend.mapping.mapping import (
    Mapping,
    Temporal,
    Storage,
    Compute as MappingCompute,
)
from accelforge.frontend.renames import Rename, RenameList, Renames
from accelforge.frontend.config import Config
from accelforge.frontend.variables import Variables

_REPO_ROOT = Path(__file__).parent.parent.parent
EXAMPLES_DIR = _REPO_ROOT / "examples"


def _yaml_spec(yaml_text: str, **jinja_parse_data) -> Spec:
    """Helper: write YAML text to a temp file and load it as a Spec."""
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".yaml", delete=False, dir=str(_REPO_ROOT)
    ) as f:
        f.write(yaml_text)
        f.flush()
        try:
            return Spec.from_yaml(f.name, jinja_parse_data=jinja_parse_data)
        finally:
            os.unlink(f.name)


# ============================================================================
# 1. Workload.get_iteration_space_shape_isl_string
# ============================================================================


class TestGetIterationSpaceShapeIslString(unittest.TestCase):
    """Test Workload.get_iteration_space_shape_isl_string."""

    def _simple_workload(self):
        return Workload(
            rank_sizes={"M": 16, "K": 32, "N": 64},
            einsums=[
                Einsum(
                    name="Matmul",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m", "K": "k"}),
                        TensorAccess(name="B", projection={"K": "k", "N": "n"}),
                        TensorAccess(
                            name="C",
                            projection={"M": "m", "N": "n"},
                            output=True,
                        ),
                    ],
                )
            ],
        )

    def test_basic_isl_string(self):
        wl = self._simple_workload()
        isl_str = wl.get_iteration_space_shape_isl_string("Matmul")
        self.assertIsInstance(isl_str, str)
        # Should contain bounds for rank variables
        self.assertIn("m", isl_str)
        self.assertIn("k", isl_str)
        self.assertIn("n", isl_str)

    def test_isl_string_contains_bounds(self):
        wl = self._simple_workload()
        isl_str = wl.get_iteration_space_shape_isl_string("Matmul")
        # Should contain upper bounds from rank_sizes
        self.assertIn("16", isl_str)
        self.assertIn("32", isl_str)
        self.assertIn("64", isl_str)

    def test_isl_string_format(self):
        """ISL string should use '0 <= var < bound' format joined by 'and'."""
        wl = self._simple_workload()
        isl_str = wl.get_iteration_space_shape_isl_string("Matmul")
        self.assertIn("0 <=", isl_str)
        self.assertIn("and", isl_str)

    def test_with_iteration_space_shape(self):
        """Workload-level iteration_space_shape is incorporated."""
        wl = Workload(
            rank_sizes={"M": 16, "K": 32},
            iteration_space_shape={"m": "0 <= m < 16"},
            einsums=[
                Einsum(
                    name="E",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m", "K": "k"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                )
            ],
        )
        isl_str = wl.get_iteration_space_shape_isl_string("E")
        self.assertIn("0 <= m < 16", isl_str)

    def test_with_einsum_rank_sizes(self):
        """Per-Einsum rank_sizes override workload-level ones."""
        wl = Workload(
            rank_sizes={"M": 16},
            einsums=[
                Einsum(
                    name="E",
                    rank_sizes={"K": 99},
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m", "K": "k"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                )
            ],
        )
        isl_str = wl.get_iteration_space_shape_isl_string("E")
        self.assertIn("99", isl_str)
        self.assertIn("16", isl_str)

    def test_from_yaml_three_matmuls(self):
        yaml_path = EXAMPLES_DIR / "workloads" / "three_matmuls_annotated.yaml"
        if not yaml_path.exists():
            self.skipTest("YAML not found")
        spec = Spec.from_yaml(yaml_path)
        # Einsum names in three_matmuls_annotated are Matmul1, Matmul2, Matmul3
        isl_str = spec.workload.get_iteration_space_shape_isl_string("Matmul1")
        self.assertIsInstance(isl_str, str)
        self.assertIn("and", isl_str)


# ============================================================================
# 2. Workload.n_instances / Einsum.n_instances
# ============================================================================


class TestNInstances(unittest.TestCase):
    """Test n_instances field on Workload and Einsum."""

    def test_einsum_n_instances_default(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        self.assertEqual(e.n_instances, 1)

    def test_einsum_n_instances_custom(self):
        e = Einsum(
            name="E",
            n_instances=4,
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        self.assertEqual(e.n_instances, 4)

    def test_workload_n_instances_default(self):
        wl = Workload()
        self.assertEqual(wl.n_instances, 1)

    def test_workload_n_instances_custom(self):
        wl = Workload(n_instances=8)
        self.assertEqual(wl.n_instances, 8)


# ============================================================================
# 3. Tensors.back, tile_shape, no_refetch_from_above, tensor_order_options
# ============================================================================


class TestTensorsAdvancedFields(unittest.TestCase):
    """Test advanced Tensors fields."""

    def test_back_default(self):
        t = Tensors(keep="All")
        self.assertEqual(t.back, "Nothing")

    def test_back_custom(self):
        t = Tensors(keep="All", back="Inputs")
        self.assertEqual(t.back, "Inputs")

    def test_tile_shape_default_empty(self):
        t = Tensors(keep="All")
        self.assertEqual(len(t.tile_shape), 0)

    def test_tile_shape_with_comparisons(self):
        t = Tensors(
            keep="All",
            tile_shape=[
                {"expression": "~m", "operator": "<=", "value": 128},
            ],
        )
        self.assertEqual(len(t.tile_shape), 1)
        self.assertIsInstance(t.tile_shape[0], Comparison)
        self.assertEqual(t.tile_shape[0].value, 128)

    def test_no_refetch_from_above_default(self):
        t = Tensors(keep="All")
        self.assertEqual(t.no_refetch_from_above, "~All")

    def test_no_refetch_from_above_custom(self):
        t = Tensors(keep="All", no_refetch_from_above="Inputs")
        self.assertEqual(t.no_refetch_from_above, "Inputs")

    def test_tensor_order_options_default_empty(self):
        t = Tensors(keep="All")
        self.assertEqual(len(t.tensor_order_options), 0)

    def test_tensor_order_options_custom(self):
        t = Tensors(
            keep="All",
            tensor_order_options=[["input | output", "weight"]],
        )
        self.assertEqual(len(t.tensor_order_options), 1)
        self.assertEqual(len(t.tensor_order_options[0]), 2)

    def test_force_memory_hierarchy_order_default(self):
        t = Tensors(keep="All")
        self.assertTrue(t.force_memory_hierarchy_order)

    def test_force_memory_hierarchy_order_false(self):
        t = Tensors(keep="All", force_memory_hierarchy_order=False)
        self.assertFalse(t.force_memory_hierarchy_order)

    def test_on_memory_from_yaml(self):
        """Test advanced Tensors fields survive YAML parsing."""
        spec = _yaml_spec("""\
arch:
  nodes:
  - !Memory
    name: Mem
    size: inf
    leak_power: 0
    area: 0
    tensors:
      keep: All
      back: Inputs
      no_refetch_from_above: Outputs
      tile_shape:
      - {expression: "~m", operator: "<=", value: 64}
    actions:
    - {name: read, energy: 1, latency: 0}
    - {name: write, energy: 1, latency: 0}
  - !Compute
    name: MAC
    leak_power: 0
    area: 0
    actions:
    - {name: compute, energy: 1, latency: 1}
""")
        mem = spec.arch.find("Mem")
        self.assertEqual(mem.tensors.keep, "All")
        self.assertEqual(mem.tensors.back, "Inputs")
        self.assertEqual(mem.tensors.no_refetch_from_above, "Outputs")
        self.assertEqual(len(mem.tensors.tile_shape), 1)
        self.assertEqual(mem.tensors.tile_shape[0].value, 64)


# ============================================================================
# 4. Arch.total_area, total_leak_power, per_component_total_area/leak
# ============================================================================


class TestArchTotalAreaLeakPower(unittest.TestCase):
    """Test Arch.total_area and related properties."""

    def _make_arch_with_totals(self):
        """Build a Spec, calculate area/leak, and return it."""
        spec = _yaml_spec("""\
arch:
  nodes:
  - !Memory
    name: MainMemory
    size: inf
    leak_power: 0.5
    area: 100
    tensors: {keep: All}
    actions:
    - {name: read, energy: 1, latency: 0}
    - {name: write, energy: 1, latency: 0}
  - !Compute
    name: MAC
    leak_power: 0.1
    area: 10
    actions:
    - {name: compute, energy: 1, latency: 1}

workload:
  rank_sizes: {M: 16}
  bits_per_value: {All: 8}
  einsums:
  - name: E
    tensor_accesses:
    - {name: A, projection: [m]}
    - {name: B, projection: [m], output: true}
""")
        return spec.calculate_component_area_energy_latency_leak()

    def test_per_component_total_area(self):
        spec = self._make_arch_with_totals()
        area = spec.arch.per_component_total_area
        self.assertIn("MainMemory", area)
        self.assertIn("MAC", area)
        self.assertEqual(area["MainMemory"], 100)
        self.assertEqual(area["MAC"], 10)

    def test_per_component_total_leak_power(self):
        spec = self._make_arch_with_totals()
        leak = spec.arch.per_component_total_leak_power
        self.assertIn("MainMemory", leak)
        self.assertIn("MAC", leak)
        self.assertAlmostEqual(leak["MainMemory"], 0.5)
        self.assertAlmostEqual(leak["MAC"], 0.1)

    def test_total_area(self):
        spec = self._make_arch_with_totals()
        self.assertAlmostEqual(spec.arch.total_area, 110)

    def test_total_leak_power(self):
        spec = self._make_arch_with_totals()
        self.assertAlmostEqual(spec.arch.total_leak_power, 0.6)

    def test_raises_without_calculation(self):
        """Accessing total_area before calculate raises ValueError."""
        spec = _yaml_spec("""\
arch:
  nodes:
  - !Memory
    name: Mem
    size: inf
    leak_power: 0
    area: 0
    actions:
    - {name: read, energy: 1, latency: 0}
    - {name: write, energy: 1, latency: 0}
  - !Compute
    name: MAC
    leak_power: 0
    area: 0
    actions:
    - {name: compute, energy: 1, latency: 1}
""")
        # total_area depends on total_area being set per-component
        # which requires calculate_component_area_energy_latency_leak
        with self.assertRaises((ValueError, TypeError)):
            _ = spec.arch.total_area


# ============================================================================
# 5. Spec.calculate_component_area_energy_latency_leak
# ============================================================================


class TestCalculateComponentAreaEnergyLatencyLeak(unittest.TestCase):
    """Test the full calculate_component_area_energy_latency_leak method."""

    def _make_spec(self):
        return _yaml_spec("""\
arch:
  nodes:
  - !Memory
    name: MainMemory
    size: inf
    leak_power: 0.5
    area: 100
    tensors: {keep: All}
    actions:
    - {name: read, energy: 2.0, latency: 10}
    - {name: write, energy: 3.0, latency: 15}
  - !Compute
    name: MAC
    leak_power: 0.1
    area: 10
    actions:
    - {name: compute, energy: 0.5, latency: 1}

workload:
  rank_sizes: {M: 16}
  bits_per_value: {All: 8}
  einsums:
  - name: E
    tensor_accesses:
    - {name: A, projection: [m]}
    - {name: B, projection: [m], output: true}
""")

    def test_returns_spec(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak()
        self.assertIsInstance(result, Spec)

    def test_area_populated(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak()
        mm = result.arch.find("MainMemory")
        self.assertEqual(mm.area, 100)
        self.assertIsNotNone(mm.total_area)

    def test_energy_populated(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak()
        mm = result.arch.find("MainMemory")
        read_action = [a for a in mm.actions if a.name == "read"][0]
        self.assertEqual(read_action.energy, 2.0)

    def test_latency_populated(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak()
        mm = result.arch.find("MainMemory")
        read_action = [a for a in mm.actions if a.name == "read"][0]
        self.assertEqual(read_action.latency, 10)

    def test_leak_power_populated(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak()
        mm = result.arch.find("MainMemory")
        self.assertEqual(mm.leak_power, 0.5)

    def test_selective_area_only(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak(
            area=True, energy=False, latency=False, leak=False
        )
        mm = result.arch.find("MainMemory")
        self.assertIsNotNone(mm.total_area)

    def test_noop_if_all_false(self):
        spec = self._make_spec()
        result = spec.calculate_component_area_energy_latency_leak(
            area=False, energy=False, latency=False, leak=False
        )
        self.assertIsInstance(result, Spec)

    def test_with_fanout_multiplies_area(self):
        """A spatial fanout should multiply area."""
        spec = _yaml_spec("""\
arch:
  nodes:
  - !Memory
    name: MainMemory
    size: inf
    leak_power: 0
    area: 0
    tensors: {keep: All}
    actions:
    - {name: read, energy: 1, latency: 0}
    - {name: write, energy: 1, latency: 0}
  - !Container
    name: Array
    spatial:
    - {name: X, fanout: 4}
  - !Compute
    name: MAC
    leak_power: 0.1
    area: 10
    actions:
    - {name: compute, energy: 1, latency: 1}

workload:
  rank_sizes: {M: 16}
  bits_per_value: {All: 8}
  einsums:
  - name: E
    tensor_accesses:
    - {name: A, projection: [m]}
    - {name: B, projection: [m], output: true}
""")
        result = spec.calculate_component_area_energy_latency_leak()
        mac = result.arch.find("MAC")
        # total_area should be area * fanout = 10 * 4
        self.assertEqual(mac.total_area, 40)


# ============================================================================
# 6. Action.latency_scale, Component.latency_scale
# ============================================================================


class TestLatencyScale(unittest.TestCase):
    """Test latency_scale field on Action and Component."""

    def test_action_latency_scale_default(self):
        a = Action(name="read", energy=1, latency=5)
        self.assertEqual(a.latency_scale, 1)

    def test_action_latency_scale_custom(self):
        a = Action(name="read", energy=1, latency=5, latency_scale=2.0)
        self.assertEqual(a.latency_scale, 2.0)

    def test_component_latency_scale_default(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        self.assertEqual(c.latency_scale, 1)

    def test_component_latency_scale_custom(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            latency_scale=0.5,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        self.assertEqual(c.latency_scale, 0.5)

    def test_component_total_latency_default(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        # Default total_latency is an expression
        self.assertIsInstance(c.total_latency, str)
        self.assertIn("sum", c.total_latency)


# ============================================================================
# 7. TensorHolder.bits_per_value_scale, bits_per_action
# ============================================================================


class TestTensorHolderFields(unittest.TestCase):
    """Test TensorHolder-specific fields."""

    def test_bits_per_value_scale_default(self):
        m = Memory(
            name="Mem",
            size=1024,
            leak_power=0,
            area=0,
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        self.assertEqual(m.bits_per_value_scale, {"All": 1})

    def test_bits_per_value_scale_custom(self):
        m = Memory(
            name="Mem",
            size=1024,
            leak_power=0,
            area=0,
            bits_per_value_scale={"All": 2},
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        self.assertEqual(m.bits_per_value_scale, {"All": 2})

    def test_bits_per_action_default_none(self):
        """TensorHolder-level bits_per_action defaults to None."""
        m = Memory(
            name="Mem",
            size=1024,
            leak_power=0,
            area=0,
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        self.assertIsNone(m.bits_per_action)

    def test_bits_per_action_custom(self):
        m = Memory(
            name="Mem",
            size=1024,
            leak_power=0,
            area=0,
            bits_per_action=32,
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        self.assertEqual(m.bits_per_action, 32)

    def test_tensor_holder_action_bits_per_action_default_is_expression(self):
        """TensorHolderAction.bits_per_action is an expression by default."""
        a = TensorHolderAction(name="read", energy=1, latency=0)
        self.assertIsInstance(a.bits_per_action, str)

    def test_tensor_holder_action_bits_per_action_custom(self):
        a = TensorHolderAction(name="read", energy=1, latency=0, bits_per_action=64)
        self.assertEqual(a.bits_per_action, 64)

    def test_toll_bits_per_value_scale(self):
        t = Toll(
            name="Q",
            direction="down",
            leak_power=0,
            area=0,
            bits_per_value_scale={"All": 0.5},
        )
        self.assertEqual(t.bits_per_value_scale, {"All": 0.5})


# ============================================================================
# 8. Spec.to_yaml() round-trip
# ============================================================================


class TestToYaml(unittest.TestCase):
    """Test to_yaml serialization.

    Note: Spec.to_yaml() fails if the Metrics enum is present (ruamel can't
    serialize it). So we test to_yaml on individual sub-models instead.
    """

    def test_workload_to_yaml_returns_string(self):
        wl = Workload(
            rank_sizes={"M": 16},
            einsums=[
                Einsum(
                    name="E",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                )
            ],
        )
        yaml_str = wl.to_yaml()
        self.assertIsInstance(yaml_str, str)
        self.assertIn("E", yaml_str)
        self.assertIn("16", yaml_str)

    def test_workload_to_yaml_file(self):
        wl = Workload(
            rank_sizes={"M": 42},
            einsums=[
                Einsum(
                    name="E",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                )
            ],
        )
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            wl.to_yaml(f.name)
        try:
            with open(f.name) as fh:
                content = fh.read()
            self.assertIn("42", content)
        finally:
            os.unlink(f.name)

    def test_round_trip_workload(self):
        """Serialize a Workload to YAML, reload as dict, verify fields."""
        orig = Workload(
            rank_sizes={"M": 16, "K": 32},
            einsums=[
                Einsum(
                    name="Matmul",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m", "K": "k"}),
                        TensorAccess(name="B", projection={"K": "k"}, output=True),
                    ],
                )
            ],
        )
        yaml_str = orig.to_yaml()
        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".yaml", delete=False, dir=str(_REPO_ROOT)
        ) as f:
            f.write("workload:\n")
            # Indent each line of the YAML string
            for line in yaml_str.splitlines():
                f.write(f"  {line}\n")
            f.flush()
            try:
                reloaded = Spec.from_yaml(f.name)
            finally:
                os.unlink(f.name)

        self.assertEqual(len(reloaded.workload.einsums), 1)
        self.assertEqual(reloaded.workload.einsums[0].name, "Matmul")

    def test_config_to_yaml(self):
        c = Config()
        yaml_str = c.to_yaml()
        self.assertIsInstance(yaml_str, str)

    def test_variables_to_yaml(self):
        v = Variables(x=42, y=99)
        yaml_str = v.to_yaml()
        self.assertIsInstance(yaml_str, str)
        self.assertIn("42", yaml_str)
        self.assertIn("99", yaml_str)

    def test_einsum_to_yaml(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        yaml_str = e.to_yaml()
        self.assertIsInstance(yaml_str, str)
        self.assertIn("E", yaml_str)


# ============================================================================
# 9. EvalableModel utility methods
# ============================================================================


class TestAllFieldsDefault(unittest.TestCase):
    """Test EvalableModel.all_fields_default."""

    def test_default_spec_all_fields_default(self):
        spec = Spec()
        self.assertTrue(spec.all_fields_default())

    def test_non_default_spec(self):
        spec = Spec(variables=Variables(x=42))
        self.assertFalse(spec.all_fields_default())

    def test_default_config(self):
        c = Config()
        self.assertTrue(c.all_fields_default())

    def test_default_workload(self):
        wl = Workload()
        self.assertTrue(wl.all_fields_default())


class TestModelDumpNonNone(unittest.TestCase):
    """Test model_dump_non_none."""

    def test_excludes_none_values(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        dump = c.model_dump_non_none()
        for k, v in dump.items():
            self.assertIsNotNone(v, f"key {k} should not be None")

    def test_includes_non_none_values(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        dump = c.model_dump_non_none()
        self.assertEqual(dump["name"], "MAC")
        self.assertEqual(dump["area"], 0)


class TestShallowModelDump(unittest.TestCase):
    """Test shallow_model_dump."""

    def test_returns_dict(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        dump = c.shallow_model_dump()
        self.assertIsInstance(dump, dict)
        self.assertIn("name", dump)

    def test_excludes_none_by_default(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        dump = c.shallow_model_dump()
        for k, v in dump.items():
            self.assertIsNotNone(v, f"key {k} should not be None")

    def test_include_none(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        dump = c.shallow_model_dump(include_None=True)
        # Should include fields that are None
        self.assertIsInstance(dump, dict)
        self.assertIn("name", dump)


class TestContainsAndGetItem(unittest.TestCase):
    """Test __contains__ and __getitem__ on EvalableModel."""

    def test_contains_existing_field(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        self.assertIn("name", e)
        self.assertIn("tensor_accesses", e)

    def test_not_contains_nonexistent(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        self.assertNotIn("nonexistent_field", e)

    def test_getitem(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        self.assertEqual(e["name"], "E")

    def test_getitem_missing_raises_keyerror(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        with self.assertRaises(KeyError):
            _ = e["nonexistent"]

    def test_setitem(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        e["n_instances"] = 5
        self.assertEqual(e.n_instances, 5)


class TestEvalableListContains(unittest.TestCase):
    """Test __contains__ and __getitem__ on EvalableList (name-based lookup)."""

    def test_einsum_list_contains_by_name(self):
        wl = Workload(
            einsums=[
                Einsum(
                    name="E1",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                ),
                Einsum(
                    name="E2",
                    tensor_accesses=[
                        TensorAccess(name="C", projection={"N": "n"}),
                        TensorAccess(name="D", projection={"N": "n"}, output=True),
                    ],
                ),
            ],
        )
        self.assertIn("E1", wl.einsums)
        self.assertIn("E2", wl.einsums)
        self.assertNotIn("E3", wl.einsums)

    def test_einsum_list_getitem_by_name(self):
        wl = Workload(
            einsums=[
                Einsum(
                    name="E1",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                ),
            ],
        )
        e = wl.einsums["E1"]
        self.assertEqual(e.name, "E1")

    def test_tensor_access_list_getitem_by_name(self):
        e = Einsum(
            name="E",
            tensor_accesses=[
                TensorAccess(name="A", projection={"M": "m"}),
                TensorAccess(name="B", projection={"M": "m"}, output=True),
            ],
        )
        a = e.tensor_accesses["A"]
        self.assertEqual(a.name, "A")


# ============================================================================
# Extra: Component.component_class
# ============================================================================


class TestComponentClass(unittest.TestCase):
    """Test Component.component_class field."""

    def test_default_none(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        self.assertIsNone(c.component_class)

    def test_custom_class(self):
        c = Compute(
            name="MAC",
            leak_power=0,
            area=0,
            component_class="MyCustomMAC",
            actions=[{"name": "compute", "energy": 1, "latency": 1}],
        )
        self.assertEqual(c.component_class, "MyCustomMAC")

    def test_get_component_class_raises_when_none(self):
        """If component_class is None, get_component_class raises EvaluationError."""
        from accelforge.util.exceptions import EvaluationError

        m = Memory(
            name="SRAM_32KB",
            size=1024,
            leak_power=0,
            area=0,
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        with self.assertRaises(EvaluationError):
            m.get_component_class()

    def test_get_component_class_when_set(self):
        """If component_class is set, get_component_class returns it."""
        m = Memory(
            name="SRAM",
            size=1024,
            leak_power=0,
            area=0,
            component_class="smartbuffer_SRAM",
            actions=[
                {"name": "read", "energy": 1, "latency": 0},
                {"name": "write", "energy": 1, "latency": 0},
            ],
        )
        cc = m.get_component_class()
        self.assertEqual(cc, "smartbuffer_SRAM")


# ============================================================================
# Extra: Workload.empty_renames, Einsum.empty_renames
# ============================================================================


class TestEmptyRenames(unittest.TestCase):
    """Test the empty_renames static/method."""

    def test_einsum_empty_renames_keys(self):
        renames = Einsum.empty_renames()
        self.assertIsInstance(renames, dict)
        # The static method returns: All, Tensors, Nothing, Inputs, Outputs
        expected_keys = {"All", "Inputs", "Outputs", "Nothing", "Tensors"}
        for key in expected_keys:
            self.assertIn(key, renames)

    def test_einsum_empty_renames_are_invertible_sets(self):
        from accelforge.util._setexpressions import InvertibleSet

        renames = Einsum.empty_renames()
        for key, val in renames.items():
            self.assertIsInstance(val, InvertibleSet, f"{key} should be InvertibleSet")

    def test_workload_empty_renames(self):
        """Workload.empty_renames is a plain method (not a property)."""
        wl = Workload(
            einsums=[
                Einsum(
                    name="E",
                    tensor_accesses=[
                        TensorAccess(name="A", projection={"M": "m"}),
                        TensorAccess(name="B", projection={"M": "m"}, output=True),
                    ],
                ),
            ],
        )
        renames = wl.empty_renames()
        self.assertIsInstance(renames, dict)
        for key in ("All", "Inputs", "Outputs", "Nothing"):
            self.assertIn(key, renames)


# ============================================================================
# Extra: Shape.rank_variables
# ============================================================================


class TestShapeRankVariables(unittest.TestCase):
    """Test Shape.rank_variables property."""

    def test_empty_shape(self):
        s = Shape()
        self.assertEqual(s.rank_variables, set())

    def test_shape_with_expressions(self):
        s = Shape(["0 <= m < 16", "0 <= k < 32"])
        rv = s.rank_variables
        self.assertIn("m", rv)
        self.assertIn("k", rv)


# ============================================================================
# Extra: Config.expression_custom_functions
# ============================================================================


class TestConfigExpressionCustomFunctions(unittest.TestCase):
    """Test that Config.expression_custom_functions stores the list."""

    def test_default_empty(self):
        c = Config()
        self.assertEqual(len(c.expression_custom_functions), 0)

    def test_from_yaml_with_paths(self):
        spec = _yaml_spec("""\
config:
  expression_custom_functions: []
  use_installed_component_models: false
  component_models: []
""")
        self.assertEqual(len(spec.config.expression_custom_functions), 0)
        self.assertFalse(spec.config.use_installed_component_models)


if __name__ == "__main__":
    unittest.main()



from collections.abc import Sequence
from dataclasses import dataclass, field
import logging
from numbers import Number
from typing import Any, Callable
from uuid import UUID, uuid4

import accelforge.frontend.arch as arch
from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Mapping,
)
from accelforge.frontend.spec import Spec
from accelforge.frontend._workload_isl._symbolic import Relevant, PartiallyRelevant
from accelforge.frontend.workload import (
    Einsum,
    EinsumName,
    RankVariable,
    SymbolTable,
    TensorName,
    Workload,
    Rank,
)

from accelforge.frontend.mapper import Metrics
from accelforge.mapper.FFM._join_pmappings.compatibility import (
    Compatibility,
)
from accelforge.mapper.FFM._make_pmappings.contraints.constraints import (
    MappingConstraints,
    _ConstraintLambda,
)
from accelforge.util.parallel import _expfmt
from accelforge.util._itertools import first
from accelforge.frontend.mapping import Reservation as ReservationNode


@dataclass
class Job:
    spec_one_einsum: Spec | None
    metrics: Metrics
    objective_tolerance: float
    rank_variable_bounds: dict[RankVariable, int]
    workload_n_einsums: int
    resource_usage_tolerance: float
    objective_tolerance: float

    job_id: UUID = field(default_factory=uuid4)

    stride_and_halo: (
        dict[
            tuple[EinsumName, TensorName],
            dict[tuple[Rank, RankVariable], tuple[int, int]],
        ]
        | None
    ) = None
    mapping: Mapping | None = None
    constraints: MappingConstraints | None = None
    fusable_tensors: set[TensorName] | None = None
    flattened_arch: list[arch.Leaf] | None = None

    einsum_name: EinsumName | None = None
    """If the Job is for a single einsum, this is the einsum name."""

    compatibility: Compatibility | None = None
    memories_track_all: list[str] | None = None
    memories_track_pmappings_only: list[str] | None = None
    ignored_resources: set[str] | None = None
    time_limit: float | int = float("inf")
    memory_limit: float | int = float("inf")
    messages: list[str] = field(default_factory=list)
    pmapping_keep_rates: dict[str, float] = field(default_factory=dict)
    tensor_to_relevancy: (
        dict[TensorName, dict[RankVariable, Relevant | PartiallyRelevant]] | None
    ) = None

    n_total_pmappings: int = 1
    n_valid_pmappings: int = 1
    n_evaluated_pmappings: int = 0

    symbol_table: SymbolTable | None = None

    initial_delta_choices: dict[RankVariable, frozenset[int]] | None = None

    ranks_with_tile_pattern: set[Rank] | None = None

    intermediate_tensors: set[TensorName] | None = None

    @property
    def einsum(self) -> Einsum:
        return self.spec_one_einsum.workload.einsums[self.einsum_name]

    @property
    def is_copy_operation(self) -> bool:
        return self.spec_one_einsum.workload.einsums[self.einsum_name].is_copy_operation

    @classmethod
    def make_job(
        cls,
        **kwargs,
    ) -> "Job":
        defaults = {
            "spec": None,
            "mapping": None,
            "workload": None,
            "architecture": None,
        }
        kwargs = {**defaults, **kwargs}
        return cls(**kwargs)

    def pretty_str(self) -> str:
        constraints = self.constraints.get_all_constraints()
        node2constraints: dict[int, list[_ConstraintLambda]] = {}
        for constraint in constraints:
            for target_index in constraint._target_node_indices:
                l = node2constraints.setdefault(target_index, [])
                l.append(constraint)

        # Reservations are added after mapping generation so it messes up the indexing
        mapping = [n for n in self.mapping.nodes if not isinstance(n, ReservationNode)]

        s = ""
        s += "=" * 80 + "\n"
        s += f"Mapper job with ID {self.job_id}\n"
        s += f"Einsum name: {self.einsum_name}\n"
        s += f"Rank variable bounds: {self.rank_variable_bounds}\n"
        s += f"Compute node name: {self.flattened_arch[-1].name}\n"
        s += f"Mapping:\n"
        for i, node in enumerate(mapping):
            cur_constraints = sorted(
                constraints.index(c) for c in node2constraints.get(i, [])
            )
            s += f"\t{i} {node.compact_str()} constrained by {cur_constraints}\n"
        s += self.constraints.pretty_str()
        s += f"Messages:\n"
        for m in self.messages:
            s += f"\t{m}\n"

        s += f"Total pmappings: {self.n_total_pmappings}\n"
        s += f"Valid pmappings: {self.n_valid_pmappings}\n"
        s += f"One in {_expfmt(self.n_total_pmappings / self.n_valid_pmappings)} pmappings is valid\n"
        s += f"Number of pmappings evaluated: {self.n_evaluated_pmappings}\n"
        s += f"One in {_expfmt(self.n_evaluated_pmappings / self.n_total_pmappings)} pmappings was evaluated\n"
        s += f"Pmapping elimination reasons:\n"
        for cause, keep_rate in self.pmapping_keep_rates.items():
            s += f"\t{cause} kept one in {_expfmt(1/keep_rate)} pmappings\n"
        s += "=" * 80 + "\n"
        return s

    def set_total_pmappings(self, n_pmappings: int):
        self.n_total_pmappings = n_pmappings

    def log_porp_pmappings_kept(
        self,
        cause: str,
        porp_kept: float,
        out_of: int = None,
    ):
        if out_of is not None:
            n_kept = porp_kept * out_of + (self.n_total_pmappings - out_of)
            porp_kept = n_kept / self.n_total_pmappings

        if any(x == 0 for x in self.pmapping_keep_rates.values()):
            return

        self.pmapping_keep_rates.setdefault(cause, 1)
        self.pmapping_keep_rates[cause] *= porp_kept

    def log_message(self, message: str):
        self.messages.append(message)
        logging.info(message)

    def __copy__(self) -> "Job":
        new = self.__class__(**self.__dict__)
        new.messages = self.messages.copy()
        new.pmapping_keep_rates = self.pmapping_keep_rates.copy()
        return new


class SameSpecJobs(list[Job]):
    @property
    def spec(self) -> Spec:
        return first(self).spec

    @property
    def rank_variable_bounds(self) -> dict[RankVariable, int]:
        return first(self).rank_variable_bounds

    @property
    def metrics(self) -> Metrics:
        return first(self).metrics


class SameEinsumJobs(SameSpecJobs):
    def check_invariance(self):
        all_einsums = oset(job.einsum_name for job in self)
        if len(all_einsums) > 1:
            raise RuntimeError("broken invariance: not all Einsums are equal.")

    @property
    def fusable_tensors(self) -> set[TensorName]:
        return first(self).fusable_tensors

    @property
    def einsum_name(self) -> set[EinsumName]:
        return first(self).einsum_name

    @property
    def rank_variable_bounds(self) -> dict[RankVariable, int]:
        return first(self).rank_variable_bounds

    @property
    def stride_and_halo(
        self,
    ) -> dict[
        tuple[EinsumName, TensorName],
        dict[tuple[Rank, RankVariable], tuple[int, int]],
    ]:
        return first(self).stride_and_halo

    @property
    def is_copy_op(self) -> bool:
        return first(self).is_copy_operation


class SameCompatibilityJobs(SameEinsumJobs):
    """Jobs with the same compatibility before tile shape exploration."""

    def check_invariance(self):
        all_compatibilities = oset(job.compatibility for job in self)
        if len(all_compatibilities) > 1:
            raise RuntimeError(
                "broken invariance: " "not all compatibilities are equal."
            )

    @property
    def compatibility(self) -> Compatibility:
        return first(self).compatibility

    def split(self) -> list["SameCompatibilityJobs"]:
        return [SameCompatibilityJobs([j]) for j in self]



import sympy
from sympy import Max
from sympy import Min
from numbers import Number

# MaxGeqZero = Max
# MinGeqZero = Min


_nonneg_true = lambda *args, **kwargs: True
_nonneg_false = lambda *args, **kwargs: False
_ZERO = sympy.Integer(0)


def MaxGeqZero(*args):
    # Fast path covers most calls, avoids sympy
    if len(args) == 2:
        a, b = args
        for a, b in ((args[0], args[1]), (args[1], args[0])):
            a_is_numeric = isinstance(a, (int, float))
            b_is_numeric = isinstance(b, (int, float))
            if a_is_numeric and b_is_numeric:
                return max(a, b, 0)
            if a_is_numeric and a == 0:
                if isinstance(b, sympy.core.numbers.Number):
                    return max(float(b), 0)
                if isinstance(b, Max) and _ZERO in b.args:
                    return b
                x = Max(b, 0)
                if isinstance(x, Max):
                    x._eval_is_nonnegative = _nonneg_true
                    x._eval_is_negative = _nonneg_false
                return x

    non_max_args = [a for a in args if not isinstance(a, Max)]
    max_args = [a for a in args if isinstance(a, Max)]
    for a in max_args:
        non_max_args.extend(a.args)

    if 0 not in non_max_args:
        non_max_args.append(0)

    x = Max(*non_max_args)
    if isinstance(x, Max):
        x._eval_is_nonnegative = _nonneg_true
        x._eval_is_negative = _nonneg_false
    return x


def MinGeqZero(*args):
    if len(args) == 2:
        a, b = args
        if isinstance(a, (int, float)) and isinstance(b, (int, float)):
            return min(a, b)

    non_min_args = [a for a in args if not isinstance(a, Min)]
    min_args = [a for a in args if isinstance(a, Min)]
    for a in min_args:
        non_min_args.extend(a.args)
    x = Min(*non_min_args)
    if isinstance(x, Min):
        x._eval_is_nonnegative = _nonneg_true
        x._eval_is_negative = _nonneg_false
    return x



from unittest import TestCase

import accelforge as af


class TestParsing(TestCase):
    def test_hierarchical(self):
        spec = af.Spec.from_yaml(
            af.examples.arches.networked.hierarchical,
        )
        self.assertIn("PeArray", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["PeArray"].get_fanout(), 1)
        self.assertIn("Scratchpad", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["Scratchpad"].get_fanout(), 4)
        self.assertIn("MacArray", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["MacArray"].get_fanout(), 1)

        try:
            spec = spec.calculate_component_area_energy_latency_leak()
        except af.EvaluationError as e:
            self.fail(e.message)

    def test_flat(self):
        spec = af.Spec.from_yaml(
            af.examples.arches.networked.flat,
        )
        self.assertIn("NoC", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["NoC"].get_fanout(), 1)
        self.assertEqual(
            {n.name for n in spec.arch.get_nodes_of_type(af.spec.Leaf)},
            {
                "MainMemory",
                "GlobalBuffer",
                "NoC",
                "RowBuffer",
                "ColumnBuffer",
                "DistributedBuffer",
                "Scratchpad",
                "MAC",
            },
        )

        try:
            spec = spec.calculate_component_area_energy_latency_leak()
        except af.EvaluationError as e:
            self.fail(e.message)


class TestModel(TestCase):
    def test_hierarchical(self):
        M = 8
        KN = 8
        MAC_TILE = 2
        PE_TILE = KN // MAC_TILE
        M_TILE = 4
        BITS_PER_VALUE = 8

        spec = af.Spec.from_yaml(
            af.examples.workloads.matmuls,
            af.examples.arches.networked.hierarchical,
            af.examples.mappings.one_matmul_to_networked_hierarchical,
            jinja_parse_data={
                "N_EINSUMS": 1,
                "M": 8,
                "KN": 8,
                "MAC_TILE": MAC_TILE,
                "M_TILE": M_TILE,
            },
        )
        result = spec.evaluate_mapping()
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T0<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (KN / MAC_TILE) ** 2
            * M_TILE
            * (0.5 * MAC_TILE * (MAC_TILE - 1) + MAC_TILE * (MAC_TILE - 1))
            * BITS_PER_VALUE,
        )
        # NOTE: assuming XY routing (as defined in mapping)
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T1<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (KN / MAC_TILE) ** 2
            * M_TILE
            * (MAC_TILE * (MAC_TILE - 1) + MAC_TILE * (MAC_TILE - 1))
            * BITS_PER_VALUE,
        )
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>W0<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (KN / MAC_TILE) ** 2
            * M_TILE
            * (MAC_TILE * (MAC_TILE - 1) + MAC_TILE * (MAC_TILE - 1))
            * BITS_PER_VALUE,
        )

        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T0<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (0.5 * PE_TILE * (PE_TILE - 1) + PE_TILE * (PE_TILE - 1))
            * M_TILE
            * MAC_TILE
            * BITS_PER_VALUE,
        )
        # NOTE: assuming XY routing (as defined in mapping)
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T1<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (PE_TILE * (PE_TILE - 1) + PE_TILE * 0.5 * PE_TILE * (PE_TILE - 1))
            * M_TILE
            * MAC_TILE
            * BITS_PER_VALUE,
        )
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>W0<SEP>hops"].iloc[0],
            (M / M_TILE)
            * (
                PE_TILE * 0.5 * PE_TILE * (PE_TILE - 1)
                + PE_TILE * 0.5 * PE_TILE * (PE_TILE - 1)
            )
            * MAC_TILE**2
            * BITS_PER_VALUE,
        )


class TestMapper(TestCase):
    def test_hierarchical(self):
        M = 8
        KN = 8
        MAC_TILE = 2
        PE_TILE = KN // MAC_TILE
        M_TILE = 4
        BITS_PER_VALUE = 8

        spec = af.Spec.from_yaml(
            af.examples.workloads.matmuls,
            af.examples.arches.networked.hierarchical,
            jinja_parse_data={
                "N_EINSUMS": 1,
                "M": 8,
                "KN": 8,
                "MAC_TILE": MAC_TILE,
                "M_TILE": M_TILE,
            },
        )
        result = spec.map_workload_to_arch()
        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T0<SEP>hops"].iloc[0],
        #     (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (0.5*MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        # )
        # # NOTE: assuming XY routing (as defined in mapping)
        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T1<SEP>hops"].iloc[0],
        #     (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        # )
        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>MacArray<SEP>W0<SEP>hops"].iloc[0],
        #     (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        # )

        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T0<SEP>hops"].iloc[0],
        #     (M/M_TILE) * (0.5*PE_TILE*(PE_TILE-1) + PE_TILE*(PE_TILE-1)) * M_TILE*MAC_TILE*BITS_PER_VALUE
        # )
        # # NOTE: assuming XY routing (as defined in mapping)
        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T1<SEP>hops"].iloc[0],
        #     (M/M_TILE) * (PE_TILE*(PE_TILE-1) + PE_TILE*0.5*PE_TILE*(PE_TILE-1)) * M_TILE*MAC_TILE*BITS_PER_VALUE
        # )
        # self.assertEqual(
        #     result.data["Matmul0<SEP>action<SEP>PeArray<SEP>W0<SEP>hops"].iloc[0],
        #     (M/M_TILE) * (PE_TILE*0.5*PE_TILE*(PE_TILE-1) + PE_TILE*0.5*PE_TILE*(PE_TILE-1)) * MAC_TILE**2*BITS_PER_VALUE
        # )
